import pytest


@pytest.mark.skip(reason="Not yet implemented")
def test_ingress_result_name(tx_port, rx_port, b2b_ipv4_devices, api):
    """
    A unique name that can be used to drill down into flow results.
    The name will appears as one of the ingress_result_name options in requesting flow results and as a property in a flow result
    ingress_result_name will configure through Pattern. This test will covered following ingress_result_name within traffic items
        -Test Data:
            VLAN - priority (priority=Pattern(choice='1', ingress_result_name='VLAN priority'))
            Ipv4 - src (src=Pattern(ingress_result_name='IPv4 src'))
        - Pause Storm
            PfcPause - src (src=Pattern('00:00:fa:ce:fa:ce', ingress_result_name='PfcPause src'))
    """
    data_endpoint = DeviceTxRx(
        tx_device_names=[b2b_ipv4_devices[0].name],
        rx_device_names=[b2b_ipv4_devices[1].name],
    )

    test_dscp = Priority(Dscp(phb=Pattern(Dscp.PHB_CS7)))
    test_flow = Flow(
        name="Test Data",
        tx_rx=TxRx(data_endpoint),
        packet=[
            Header(Ethernet()),
            Header(
                Vlan(
                    priority=Pattern(
                        choice="1", ingress_result_name="VLAN priority"
                    )
                )
            ),
            Header(Ipv4(priority=test_dscp)),
        ],
        size=Size(128),
        rate=Rate("line", 50),
        duration=Duration(FixedPackets(packets=0)),
    )

    pause_endpoint = PortTxRx(
        tx_port_name=tx_port.name, rx_port_name=rx_port.name
    )
    pause = Header(
        PfcPause(
            dst=Pattern("01:80:C2:00:00:01"),
            src=Pattern(
                "00:00:fa:ce:fa:ce", ingress_result_name="PfcPause src"
            ),
            class_enable_vector=Pattern("1"),
            pause_class_0=Pattern("3"),
            pause_class_1=Pattern(Counter(start="2", step="6", count=99)),
            pause_class_2=Pattern(
                Counter(start="1", step="6", count=99, up=False)
            ),
            pause_class_3=Pattern(["6", "9", "2", "39"]),
            pause_class_4=Pattern(
                Random(min="11", max="33", step=1, seed="4", count=10)
            ),
        )
    )
    pause_flow = Flow(
        name="Pause Storm",
        tx_rx=TxRx(pause_endpoint),
        packet=[pause],
        size=Size(64),
        rate=Rate("line", value=100),
        duration=Duration(FixedPackets(packets=0)),
    )

    config = Config(
        ports=[tx_port, rx_port],
        devices=b2b_ipv4_devices,
        flows=[test_flow, pause_flow],
    )
    api.set_state(State(ConfigState(config=config, state="set")))


if __name__ == "__main__":
    pytest.main(["-s", __file__])



# import re
# import socket, struct
# from collections import namedtuple
#
#
# class ConfigureBgp(object):
#     _BGP_AS_SET_MODE = {
#         "do_not_include_as": "dontincludelocalas",
#         "include_as_seq": "includelocalasasasseq",
#         "include_as_set": "includelocalasasasset",
#         "include_as_seq_confed": "includelocalasasasseqconfederation",
#         "include_as_set_confed": "includelocalasasassetconfederation",
#         "prepend_as_to_first_segment": "prependlocalastofirstsegment",
#     }
#
#     _BGP_AS_MODE = {
#         "do_not_include_local_as": "dontincludelocalas",
#         "include_as_seq": "includelocalasasasseq",
#         "include_as_set": "includelocalasasasset",
#         "include_as_confed_seq": "includelocalasasasseqconfederation",
#         "include_as_confed_set": "includelocalasasassetconfederation",
#         "prepend_to_first_segment": "prependlocalastofirstsegment",
#     }
#
#     _BGP_SEG_TYPE = {
#         "as_seq": "asseq",
#         "as_set": "asset",
#         "as_confed_seq": "asseqconfederation",
#         "as_confed_set": "assetconfederation",
#     }
#
#     _BGP_COMMUNITY_TYPE = {
#         "manual_as_number": "manual",
#         "no_export": "noexport",
#         "no_advertised": "noadvertised",
#         "no_export_subconfed": "noexport_subconfed",
#         "llgr_stale": "llgr_stale",
#         "no_llgr": "no_llgr",
#     }
#
#     _BGP_SR_TE = {
#         "policy_type": {"ixn_attr": "policyType", "default": "ipv4"},
#         "distinguisher": {"ixn_attr": "distinguisher", "default": "1"},
#         "color": {"ixn_attr": "policyColor", "default": "101"},
#         "ipv4_endpoint": {"ixn_attr": "endPointV4", "default": "0.0.0.0"},
#         "ipv6_endpoint": {
#             "ixn_attr": "endPointV6",
#             "default": "0:0:0:0:0:0:0:0",
#         },
#     }
#
#     _SRTE_NEXT_HOP = {
#         "next_hop_mode": {
#             "ixn_attr": "setNextHop",
#             "default": "sameaslocalip",
#             "enum_map": {"local_ip": "sameaslocalip", "manual": "manually"},
#         },
#         "next_hop_address_type": {
#             "ixn_attr": "setNextHopIpType",
#             "default": "ipv4",
#         },
#         "ipv4_address": {"ixn_attr": "ipv4NextHop", "default": "0.0.0.0"},
#         "ipv6_address": {"ixn_attr": "ipv6NextHop", "default": "::"},
#     }
#
#     _SRTE_ADDPATH = {"path_id": {"ixn_attr": "addPathId", "default": "1"}}
#
#     _SRTE_AS_PATH = {
#         "override_peer_as_set_mode": {
#             "ixn_attr": "overridePeerAsSetMode",
#             "default": False,
#         },
#         "as_set_mode": {
#             "ixn_attr": "asSetMode",
#             "default": "includelocalasasasseq",
#             "enum_map": _BGP_AS_MODE,
#         },
#     }
#
#     _SRTE_ASPATH_SEGMENT = {
#         "segment_type": {
#             "ixn_attr": "segmentType",
#             "default": "asset",
#             "enum_map": _BGP_SEG_TYPE,
#         }
#     }
#
#     _REMOTE_ENDPOINT_SUB_TLV = {
#         "as_number": {"ixn_attr": "as4Number", "default": "0"},
#         "address_family": {"ixn_attr": "addressFamily", "default": "ipv4"},
#         "ipv4_address": {
#             "ixn_attr": "remoteEndpointIPv4",
#             "default": "0.0.0.0",
#         },
#         "ipv6_address": {"ixn_attr": "remoteEndpointIPv6", "default": "::"},
#     }
#
#     _PREFERENCE_SUB_TLV = {
#         "preference": {"ixn_attr": "prefValue", "default": "0"}
#     }
#
#     _BINDING_SUB_TLV = {
#         "binding_sid_type": {
#             "ixn_attr": "bindingSIDType",
#             "default": "nobinding",
#             "enum_map": {
#                 "no_binding": "nobinding",
#                 "four_octet_sid": "sid4",
#                 "ipv6_sid": "ipv6sid",
#             },
#         },
#         "four_octet_sid": {"ixn_attr": "SID4Octet", "default": "0"},
#         "bsid_as_mpls_label": {"ixn_attr": "useAsMPLSLabel", "default": False},
#         "ipv6_sid": {"ixn_attr": "IPv6SID", "default": "::"},
#         "s_flag": {"ixn_attr": "sflag", "default": False},
#         "i_flag": {"ixn_attr": "iflag", "default": False},
#         "remaining_flag_bits": {
#             "ixn_attr": "remainingBits",
#             "default": "0x01",
#         },
#     }
#
#     _ENLP_SUB_TLV = {
#         "explicit_null_label_policy": {
#             "ixn_attr": "ENLPValue",
#             "default": "4",
#             "enum_map": {
#                 "reserved_enlp": "0",
#                 "push_ipv4_enlp": "1",
#                 "push_ipv6_enlp": "2",
#                 "push_ipv4_ipv6_enlp": "3",
#                 "do_not_push_enlp": "4",
#             },
#         }
#     }
#
#     _POLICIES_SEGMENT_LIST = {
#         "segment_weight": {"ixn_attr": "weight", "default": "200"}
#     }
#
#     _SEGMENTS = {
#         "segment_type": {
#             "ixn_attr": "segmentType",
#             "default": "mplssid",
#             "enum_map": {"mpls_sid": "mplssid", "ipv6_sid": "ipv6sid"},
#         },
#         "mpls_label": {"ixn_attr": "label", "default": "16"},
#         "mpls_tc": {"ixn_attr": "trafficClass", "default": "0"},
#         "mpls_ttl": {"ixn_attr": "timeToLive", "default": "255"},
#         "v_flag": {"ixn_attr": "vflag", "default": False},
#         "remaining_flag_bits": {
#             "ixn_attr": "remainingBits",
#             "default": "0x01",
#         },
#         "ipv6_sid": {"ixn_attr": "ipv6SID", "default": "::"},
#     }
#
#     def __init__(self, ngpf):
#         self._ngpf = ngpf
#         self._api = ngpf._api
#         self.update = ngpf.update
#         self.configure_value = ngpf.configure_value
#         self.get_xpath = ngpf.get_xpath
#         self.select_node = ngpf.select_node
#         self.select_child_node = ngpf.select_child_node
#
#     def configure_bgpv4(self, ixn_parent, bgpv4, ixn_dg):
#         ixn_bgpv4 = ixn_parent.BgpIpv4Peer
#         self._api._remove(ixn_bgpv4, [bgpv4])
#         bgp_name = bgpv4.get("name")
#         name = self._api.special_char(bgp_name)
#         args = {
#             "Name": name,
#         }
#         ixn_bgpv4.find(Name="^%s$" % name)
#         if len(ixn_bgpv4) == 0:
#             ixn_bgpv4.add(**args)[-1]
#         else:
#             self.update(ixn_bgpv4, **args)
#         as_type = "internal"
#         if bgpv4.get("as_type") is not None and bgpv4.get("as_type") == "ebgp":
#             as_type = "external"
#         bgp_xpath = self.get_xpath(ixn_bgpv4.href)
#         self._api.set_ixn_cmp_object(bgpv4, ixn_bgpv4.href, bgp_xpath)
#         self.configure_value(bgp_xpath, "type", as_type)
#         as_bytes = bgpv4.get("as_number_width")
#         as_bytes_list = (
#             [as_bytes] if not isinstance(as_bytes, list) else as_bytes
#         )
#         as_number = bgpv4.get("as_number")
#         as_number_list = (
#             [as_number] if not isinstance(as_number, list) else as_number
#         )
#         for index, as_number in enumerate(as_number_list):
#             as_byte = as_bytes_list[index]
#             if as_byte == "two":
#                 self.configure_value(bgp_xpath, "localAs2Bytes", as_number)
#             elif as_byte == "four":
#                 self.configure_value(bgp_xpath, "enable4ByteAs", True)
#                 self.configure_value(bgp_xpath, "localAs4Bytes", as_number)
#             else:
#                 msg = "Please configure supported [two, four] as_number_width"
#                 raise Exception(msg)
#         dut_address = bgpv4.get("dut_address")
#         if dut_address is not None:
#             self.configure_value(bgp_xpath, "dutIp", dut_address)
#
#         as_number_set_mode = bgpv4.get("as_number_set_mode")
#         if as_number_set_mode:
#             self.configure_value(
#                 bgp_xpath,
#                 "asSetMode",
#                 as_number_set_mode,
#                 enum_map=ConfigureBgp._BGP_AS_SET_MODE,
#             )
#         # self._configure_pattern(ixn_dg.RouterData.RouterId, bgpv4.router_id)
#         advanced = bgpv4.get("advanced")
#         if advanced is not None:
#             self.configure_value(
#                 bgp_xpath, "holdTimer", advanced.get("hold_time_interval")
#             )
#             self.configure_value(
#                 bgp_xpath,
#                 "keepaliveTimer",
#                 advanced.get("keep_alive_interval"),
#             )
#             self.configure_value(bgp_xpath, "md5Key", advanced.get("md5_key"))
#             self.configure_value(
#                 bgp_xpath, "updateInterval", advanced.get("update_interval")
#             )
#             self.configure_value(bgp_xpath, "ttl", advanced.time_to_live)
#         sr_te_policies = bgpv4.get("sr_te_policies")
#         if sr_te_policies is not None:
#             self._configure_sr_te(ixn_bgpv4, bgp_xpath, sr_te_policies)
#         self._bgp_route_builder(ixn_dg, ixn_bgpv4, bgpv4)
#         return ixn_bgpv4
#
#     def _bgp_route_builder(self, ixn_dg, ixn_bgp, bgp):
#         bgpv4_routes = bgp.get("bgpv4_routes")
#         bgpv6_routes = bgp.get("bgpv6_routes")
#         if bgpv4_routes is not None and len(bgpv4_routes) > 0:
#             for route_range in bgpv4_routes:
#                 self._configure_bgpv4_route(ixn_dg, ixn_bgp, route_range)
#         if bgpv6_routes is not None and len(bgpv6_routes) > 0:
#             for route_range in bgpv6_routes:
#                 self._configure_bgpv6_route(ixn_dg, ixn_bgp, route_range)
#
#     def _configure_bgpv4_route(self, ixn_dg, ixn_bgp, route_range):
#         ixn_ng = ixn_dg.NetworkGroup
#         route_name = route_range.get("name")
#         name = self._api.special_char(route_name)
#         args = {
#             "Name": name,
#         }
#         ixn_ng.find(Name="^%s$" % name)
#         if len(ixn_ng) == 0:
#             self.stop_topology()
#             ixn_ng.add(**args)[-1]
#             ixn_pool = ixn_ng.Ipv4PrefixPools.add()
#         else:
#             self.update(ixn_ng, **args)
#             ixn_pool = ixn_ng.Ipv4PrefixPools.find()
#         ixn_pool.Connector.find().ConnectedTo = ixn_bgp.href
#         pool_infos = self.select_node(
#             ixn_pool.href,
#             children=["bgpIPRouteProperty", "bgpV6IPRouteProperty"],
#         )
#         pool_xpath = pool_infos["xpath"]
#         addresses = route_range.get("addresses")
#         route_len = len(addresses)
#         if len(addresses) > 0:
#             ixn_ng.Multiplier = route_len
#             route_addresses = RouteAddresses()
#             for address in addresses:
#                 # below properties will set to default when
#                 # route_address is instantiated
#                 route_addresses.address = address.get("address")
#                 route_addresses.step = address.get("step")
#                 route_addresses.prefix = address.get("prefix")
#                 route_addresses.count = address.get("count")
#             self.configure_value(
#                 pool_xpath, "networkAddress", route_addresses.address
#             )
#             self.configure_value(
#                 pool_xpath, "prefixAddrStep", route_addresses.step
#             )
#             self.configure_value(
#                 pool_xpath, "prefixLength", route_addresses.prefix
#             )
#             self.configure_value(
#                 pool_xpath, "numberOfAddressesAsy", route_addresses.count
#             )
#         if "bgpIPRouteProperty" in pool_infos:
#             ixn_bgp_property = ixn_pool.BgpIPRouteProperty.find()
#             property_xpath = pool_infos["bgpIPRouteProperty"][0]["xpath"]
#         else:
#             ixn_bgp_property = ixn_pool.BgpV6IPRouteProperty.find()
#             property_xpath = pool_infos["bgpV6IPRouteProperty"][0]["xpath"]
#         next_hop_address = route_range.get("next_hop_address")
#         if next_hop_address:
#             self.configure_value(
#                 property_xpath,
#                 "ipv4NextHop",
#                 next_hop_address,
#                 multiplier=route_len,
#             )
#         if route_name is not None:
#             ixn_bgp_property.Name = route_name
#             self._api.set_ixn_cmp_object(
#                 route_range, ixn_pool.href, pool_xpath, multiplier=route_len
#             )
#             self._api.set_device_encap(route_range, "ipv4")
#             self._api.set_route_objects(
#                 ixn_bgp_property, route_range, multiplier=route_len
#             )
#         advanced = route_range.get("advanced")
#         if (
#             advanced is not None
#             and advanced.get("multi_exit_discriminator") is not None
#         ):
#             self.configure_value(
#                 property_xpath, "enableMultiExitDiscriminator", True
#             )
#             self.configure_value(
#                 property_xpath,
#                 "multiExitDiscriminator",
#                 advanced.get("multi_exit_discriminator"),
#                 multiplier=route_len,
#             )
#         if advanced is not None:
#             self.configure_value(
#                 property_xpath,
#                 "origin",
#                 advanced.get("origin"),
#                 multiplier=route_len,
#             )
#         as_path = route_range.get("as_path")
#         if as_path is not None:
#             self._config_bgp_as_path(as_path, ixn_bgp_property, route_len)
#         communities = route_range.get("communities")
#         if communities:
#             self._config_bgp_community(
#                 communities, ixn_bgp_property, route_len
#             )
#
#     def configure_bgpv6(self, ixn_parent, bgpv6, ixn_dg):
#         ixn_bgpv6 = ixn_parent.BgpIpv6Peer
#         self._api._remove(ixn_bgpv6, [bgpv6])
#         bgp_name = bgpv6.get("name")
#         name = self._api.special_char(bgp_name)
#         args = {
#             "Name": name,
#         }
#         ixn_bgpv6.find(Name="^%s$" % name)
#         if len(ixn_bgpv6) == 0:
#             ixn_bgpv6.add(**args)[-1]
#         else:
#             self.update(ixn_bgpv6, **args)
#         as_type = "internal"
#         if bgpv6.get("as_type") is not None and bgpv6.get("as_type") == "ebgp":
#             as_type = "external"
#         bgp_xpath = self.get_xpath(ixn_bgpv6.href)
#         self._api.set_ixn_cmp_object(bgpv6, ixn_bgpv6.href, bgp_xpath)
#         self.configure_value(bgp_xpath, "type", as_type)
#         as_bytes = bgpv6.get("as_number_width")
#         as_bytes_list = (
#             [as_bytes] if not isinstance(as_bytes, list) else as_bytes
#         )
#         as_number = bgpv6.get("as_number")
#         as_number_list = (
#             [as_number] if not isinstance(as_number, list) else as_number
#         )
#         for index, as_number in enumerate(as_number_list):
#             as_byte = as_bytes_list[index]
#             if as_byte == "two":
#                 self.configure_value(bgp_xpath, "localAs2Bytes", as_number)
#             elif as_byte == "four":
#                 self.configure_value(bgp_xpath, "enable4ByteAs", True)
#                 self.configure_value(bgp_xpath, "localAs4Bytes", as_number)
#             else:
#                 msg = "Please configure supported [two, four] as_number_width"
#                 raise Exception(msg)
#         dut_address = bgpv6.get("dut_address")
#         if dut_address is not None:
#             self.configure_value(bgp_xpath, "dutIp", dut_address)
#         as_number_set_mode = bgpv6.get("as_number_set_mode")
#         if as_number_set_mode is not None:
#             self.configure_value(
#                 bgp_xpath,
#                 "asSetMode",
#                 as_number_set_mode,
#                 enum_map=ConfigureBgp._BGP_AS_SET_MODE,
#             )
#         # self._configure_pattern(ixn_dg.RouterData.RouterId, bgpv4.router_id)
#         advanced = bgpv6.get("advanced")
#         if advanced is not None:
#             self.configure_value(
#                 bgp_xpath, "holdTimer", advanced.get("hold_time_interval")
#             )
#             self.configure_value(
#                 bgp_xpath,
#                 "keepaliveTimer",
#                 advanced.get("keep_alive_interval"),
#             )
#             self.configure_value(bgp_xpath, "md5Key", advanced.get("md5_key"))
#             self.configure_value(
#                 bgp_xpath, "updateInterval", advanced.get("update_interval")
#             )
#             self.configure_value(
#                 bgp_xpath, "ttl", advanced.get("time_to_live")
#             )
#         sr_te_policies = bgpv6.get("sr_te_policies")
#         if sr_te_policies:
#             self._configure_sr_te(ixn_bgpv6, bgp_xpath, sr_te_policies)
#         self._bgp_route_builder(ixn_dg, ixn_bgpv6, bgpv6)
#         return ixn_bgpv6
#
#     def _configure_bgpv6_route(self, ixn_dg, ixn_bgp, route_range):
#         ixn_ng = ixn_dg.NetworkGroup
#         route_name = route_range.get("name")
#         name = self._api.special_char(route_name)
#         args = {
#             "Name": name,
#         }
#         ixn_ng.find(Name="^%s$" % name)
#         if len(ixn_ng) == 0:
#             self.stop_topology()
#             ixn_ng.add(**args)[-1]
#             ixn_pool = ixn_ng.Ipv6PrefixPools.add()
#         else:
#             self.update(ixn_ng, **args)
#             ixn_pool = ixn_ng.Ipv6PrefixPools.find()
#         ixn_pool.Connector.find().ConnectedTo = ixn_bgp.href
#         pool_infos = self.select_node(
#             ixn_pool.href,
#             children=["bgpIPRouteProperty", "bgpV6IPRouteProperty"],
#         )
#         pool_xpath = pool_infos["xpath"]
#         addresses = route_range.get("addresses")
#         route_len = len(addresses)
#         if len(addresses) > 0:
#             ixn_ng.Multiplier = route_len
#             route_addresses = RouteAddresses()
#             for address in addresses:
#                 route_addresses.address = address.get("address")
#                 route_addresses.step = address.get("step")
#                 route_addresses.prefix = address.get("prefix")
#                 route_addresses.count = address.get("count")
#             self.configure_value(
#                 pool_xpath, "networkAddress", route_addresses.address
#             )
#             self.configure_value(
#                 pool_xpath, "prefixAddrStep", route_addresses.step
#             )
#             self.configure_value(
#                 pool_xpath, "prefixLength", route_addresses.prefix
#             )
#             self.configure_value(
#                 pool_xpath, "numberOfAddressesAsy", route_addresses.count
#             )
#         if self._api.get_device_encap(ixn_dg.Name) == "ipv4":
#             ixn_bgp_property = ixn_pool.BgpIPRouteProperty.find()
#             property_xpath = pool_infos["bgpIPRouteProperty"][0]["xpath"]
#         else:
#             ixn_bgp_property = ixn_pool.BgpV6IPRouteProperty.find()
#             property_xpath = pool_infos["bgpV6IPRouteProperty"][0]["xpath"]
#         next_hop_address = route_range.get("next_hop_address")
#         if next_hop_address is not None:
#             self.configure_value(
#                 property_xpath,
#                 "ipv6NextHop",
#                 next_hop_address,
#                 multiplier=route_len,
#             )
#         if route_name is not None:
#             ixn_bgp_property.Name = route_name
#             self._api.set_ixn_cmp_object(
#                 route_range, ixn_pool.href, pool_xpath, multiplier=route_len
#             )
#             self._api.set_device_encap(route_range, "ipv6")
#             self._api.set_route_objects(
#                 ixn_bgp_property, route_range, multiplier=route_len
#             )
#         advanced = route_range.get("advanced")
#         if (
#             advanced is not None
#             and advanced.get("multi_exit_discriminator") is not None
#         ):
#             self.configure_value(
#                 property_xpath, "enableMultiExitDiscriminator", True
#             )
#             self.configure_value(
#                 property_xpath,
#                 "multiExitDiscriminator",
#                 advanced.get("multi_exit_discriminator"),
#                 multiplier=route_len,
#             )
#         if advanced is not None:
#             self.configure_value(
#                 property_xpath,
#                 "origin",
#                 advanced.get("origin"),
#                 multiplier=route_len,
#             )
#         as_path = route_range.get("as_path")
#         if as_path is not None:
#             self._config_bgp_as_path(as_path, ixn_bgp_property, route_len)
#         communities = route_range.get("communities")
#         if communities:
#             self._config_bgp_community(
#                 communities, ixn_bgp_property, route_len
#             )
#
#     def _config_bgp_as_path(self, as_path, ixn_bgp_property, multiplier):
#         as_path_segments = as_path.get("as_path_segments")
#         property_xpath = self.get_xpath(ixn_bgp_property.href)
#         as_set_mode = as_path.get("as_set_mode")
#         if as_set_mode is not None or len(as_path_segments) > 0:
#             self.configure_value(property_xpath, "enableAsPathSegments", True)
#             self.configure_value(
#                 property_xpath,
#                 "asSetMode",
#                 as_set_mode,
#                 enum_map=ConfigureBgp._BGP_AS_MODE,
#                 multiplier=multiplier,
#             )
#             self.configure_value(
#                 property_xpath,
#                 "OverridePeerAsSetMode",
#                 as_path.get("override_peer_as_set_mode"),
#                 multiplier=multiplier,
#             )
#             if len(as_path_segments) > 0:
#                 ixn_bgp_property.NoOfASPathSegmentsPerRouteRange = len(
#                     as_path_segments
#                 )
#                 ixn_segments = ixn_bgp_property.BgpAsPathSegmentList.find()
#                 for seg_index, segment in enumerate(as_path_segments):
#                     ixn_segment = ixn_segments[seg_index]
#                     ixn_segment.SegmentType.Single(
#                         ConfigureBgp._BGP_SEG_TYPE[segment.get("segment_type")]
#                     )
#                     as_numbers = segment.get("as_numbers")
#                     if as_numbers is not None:
#                         ixn_segment.NumberOfAsNumberInSegment = len(as_numbers)
#                         as_numbers_info = self.select_child_node(
#                             ixn_segment.href, "bgpAsNumberList"
#                         )
#                         for as_index, as_number in enumerate(as_numbers):
#                             as_num_xpath = as_numbers_info[as_index]["xpath"]
#                             self.configure_value(
#                                 as_num_xpath,
#                                 "asNumber",
#                                 as_number,
#                                 multiplier=multiplier,
#                             )
#
#     def _config_bgp_community(self, communities, ixn_bgp_property, multiplier):
#         if len(communities) == 0:
#             ixn_bgp_property.EnableCommunity.Single(False)
#             return
#         ixn_bgp_property.EnableCommunity.Single(True)
#         ixn_bgp_property.NoOfCommunities = len(communities)
#         communities_info = self.select_child_node(
#             ixn_bgp_property.href, "bgpCommunitiesList"
#         )
#         for index, community in enumerate(communities):
#             community_xpath = communities_info[index]["xpath"]
#             community_type = community.get("community_type")
#             if community_type is not None:
#                 self.configure_value(
#                     community_xpath,
#                     "type",
#                     community_type,
#                     enum_map=ConfigureBgp._BGP_COMMUNITY_TYPE,
#                     multiplier=multiplier,
#                 )
#             self.configure_value(
#                 community_xpath,
#                 "asNumber",
#                 community.get("as_number"),
#                 multiplier=multiplier,
#             )
#             self.configure_value(
#                 community_xpath,
#                 "lastTwoOctets",
#                 community.get("as_custom"),
#                 multiplier=multiplier,
#             )
#
#     def _configure_sr_te(self, ixn_bgp, bgp_xpath, sr_te_list):
#         if sr_te_list is None or len(sr_te_list) == 0:
#             return
#         self.configure_value(bgp_xpath, "capabilitySRTEPoliciesV4", True)
#         self.configure_value(bgp_xpath, "capabilitySRTEPoliciesV6", True)
#         ixn_bgp.NumberSRTEPolicies = len(sr_te_list)
#         if re.search("bgpIpv4Peer", ixn_bgp.href) is not None:
#             ixn_sr_te = ixn_bgp.BgpSRTEPoliciesListV4
#         else:
#             ixn_sr_te = ixn_bgp.BgpSRTEPoliciesListV6
#         sr_te_xpath = self.get_xpath(ixn_sr_te.href)
#         self._configure_attributes(
#             ConfigureBgp._BGP_SR_TE, sr_te_list, sr_te_xpath
#         )
#         next_hops = []
#         add_paths = []
#         as_paths = []
#         communities = []
#         for sr_te in sr_te_list:
#             if sr_te.get("next_hop") is not None:
#                 next_hops.append(sr_te.next_hop)
#             if sr_te.get("add_path") is not None:
#                 add_paths.append(sr_te.add_path)
#             if sr_te.get("as_path") is not None:
#                 as_paths.append(sr_te.as_path)
#             if sr_te.get("communities") is not None:
#                 communities.append(sr_te.communities)
#
#         active_list = self._process_nodes(next_hops)
#         if active_list != []:
#             self.configure_value(sr_te_xpath, "enableNextHop", active_list)
#         if any(active_list):
#             self._configure_attributes(
#                 ConfigureBgp._SRTE_NEXT_HOP, next_hops, sr_te_xpath
#             )
#
#         active_list = self._process_nodes(add_paths)
#         if active_list != []:
#             self.configure_value(sr_te_xpath, "enableAddPath", active_list)
#         if any(active_list):
#             self._configure_attributes(
#                 ConfigureBgp._SRTE_ADDPATH, add_paths, sr_te_xpath
#             )
#
#         active_list = self._process_nodes(as_paths)
#         if any(active_list):
#             self._configure_attributes(
#                 ConfigureBgp._SRTE_AS_PATH, as_paths, sr_te_xpath
#             )
#             self._configure_srte_aspath_segment(as_paths, ixn_sr_te)
#         self._configure_tlvs(ixn_sr_te, sr_te_list)
#
#     def _get_symmetric_nodes(self, parent_list, node_name):
#         NodesInfo = namedtuple(
#             "NodesInfo", ["max_len", "active_list", "symmetric_nodes"]
#         )
#         nodes_list = []
#         max_len = 0
#         for parent in parent_list:
#             nodes = getattr(parent, node_name)
#             node_len = len(nodes)
#             if node_len > max_len:
#                 max_len = node_len
#             nodes_list.append(nodes)
#         symmetric_nodes = []
#         active_list = []
#         for nodes in nodes_list:
#             if len(nodes) == max_len:
#                 for node in nodes:
#                     active_list.append(node.active)
#                     symmetric_nodes.append(node)
#             else:
#                 for index in range(0, max_len):
#                     node = nodes[0]
#                     if index < len(nodes):
#                         node = nodes[index]
#                         active_list.append(node.active)
#                         symmetric_nodes.append(node)
#                     else:
#                         active_list.append(False)
#                         symmetric_nodes.append(node)
#         return NodesInfo(max_len, active_list, symmetric_nodes)
#
#     def _get_symetric_tab_nodes(self, parent_list, node_name):
#         TabNodesInfo = namedtuple(
#             "TabNodesInfo", ["max_len", "symmetric_nodes_list", "actives_list"]
#         )
#         max_len = 0
#         symmetric_nodes_list = []
#         actives_list = []
#         is_enable = False
#         for parent in parent_list:
#             nodes = getattr(parent, node_name)
#             if nodes is None:
#                 continue
#             is_enable = True
#             node_len = len(nodes)
#             if node_len > max_len:
#                 for index in range(max_len, node_len):
#                     symmetric_nodes_list.append(
#                         [nodes[index]] * len(parent_list)
#                     )
#                     actives_list.append([False] * len(parent_list))
#                 max_len = node_len
#         if is_enable:
#             for parent_idx, parent in enumerate(parent_list):
#                 nodes = getattr(parent, node_name)
#                 for node_idx, node in enumerate(nodes):
#                     symmetric_nodes_list[node_idx][parent_idx] = node
#                     actives_list[node_idx][parent_idx] = True
#         return TabNodesInfo(max_len, symmetric_nodes_list, actives_list)
#
#     def _configure_srte_aspath_segment(self, as_paths, ixn_sr_te):
#         nodes_list_info = self._get_symetric_tab_nodes(
#             as_paths, "as_path_segments"
#         )
#         if nodes_list_info.max_len == 0:
#             return
#         ixn_sr_te.EnableAsPathSegments.Single(True)
#         ixn_sr_te.NoOfASPathSegmentsPerRouteRange = nodes_list_info.max_len
#         ixn_segments = ixn_sr_te.BgpAsPathSegmentList
#         ixn_segments.find()
#         segments_info = self.select_node(
#             ixn_sr_te.refresh().href, children=["bgpAsPathSegmentList"]
#         )
#         for seg_idx, segment in enumerate(
#             segments_info["bgpAsPathSegmentList"]
#         ):
#             segment_xpath = segment["xpath"]
#             ixn_segment = ixn_segments[seg_idx]
#             segment_nodes = nodes_list_info.symmetric_nodes_list[seg_idx]
#             self.configure_value(
#                 segment_xpath,
#                 "enableASPathSegment",
#                 nodes_list_info.actives_list[seg_idx],
#             )
#             self._configure_attributes(
#                 ConfigureBgp._SRTE_ASPATH_SEGMENT, segment_nodes, segment_xpath
#             )
#             configure_as_number = False
#             for segment_node in segment_nodes:
#                 as_numbers = getattr(segment_node, "as_numbers")
#                 if as_numbers is not None:
#                     configure_as_number = True
#                     if not isinstance(as_numbers, list):
#                         raise Exception("as_numbers must be list")
#             if configure_as_number is True:
#                 as_numbers_list = self._get_symetric_tab_nodes(
#                     segment_nodes, "as_numbers"
#                 )
#                 ixn_segment.NumberOfAsNumberInSegment = as_numbers_list.max_len
#                 numbers_info = self.select_node(
#                     ixn_segment.href, children=["bgpAsNumberList"]
#                 )
#                 for num_idx, number in enumerate(
#                     numbers_info["bgpAsNumberList"]
#                 ):
#                     number_xpath = number["xpath"]
#                     self.configure_value(
#                         number_xpath,
#                         "enableASNumber",
#                         as_numbers_list.actives_list[num_idx],
#                     )
#                     self.configure_value(
#                         number_xpath,
#                         "asNumber",
#                         as_numbers_list.symmetric_nodes_list[num_idx],
#                     )
#
#     def _configure_tlvs(self, ixn_sr_te, sr_te_list):
#         nodes_info = self._get_symmetric_nodes(sr_te_list, "tunnel_tlvs")
#         if int(nodes_info.max_len) > 2:
#             raise Exception(
#                 "Value {0} for SR TE Policy Number of Tunnel TLVs is "
#                 "greater than maximal value 2".format(nodes_info.max_len)
#             )
#         if re.search("bgpSRTEPoliciesListV4", ixn_sr_te.href) is not None:
#             ixn_sr_te.NumberOfTunnelsV4 = nodes_info.max_len
#             ixn_tunnel = ixn_sr_te.BgpSRTEPoliciesTunnelEncapsulationListV4
#         else:
#             ixn_sr_te.NumberOfTunnelsV6 = nodes_info.max_len
#             ixn_tunnel = ixn_sr_te.BgpSRTEPoliciesTunnelEncapsulationListV6
#         tunnel_xpath = self.get_xpath(ixn_tunnel.href)
#         self.configure_value(tunnel_xpath, "active", nodes_info.active_list)
#         tunnel_tlvs = nodes_info.symmetric_nodes
#
#         remote_endpoint_sub_tlv = []
#         preference_sub_tlv = []
#         binding_sub_tlv = []
#         explicit_null_label_policy_sub_tlv = []
#         for tunnel_tlv in tunnel_tlvs:
#             if tunnel_tlv.get("remote_endpoint_sub_tlv") is not None:
#                 remote_endpoint_sub_tlv.append(
#                     tunnel_tlv.remote_endpoint_sub_tlv
#                 )
#             if tunnel_tlv.get("preference_sub_tlv") is not None:
#                 preference_sub_tlv.append(tunnel_tlv.preference_sub_tlv)
#             if tunnel_tlv.get("binding_sub_tlv") is not None:
#                 binding_sub_tlv.append(tunnel_tlv.binding_sub_tlv)
#             if (
#                 tunnel_tlv.get("explicit_null_label_policy_sub_tlv")
#                 is not None
#             ):
#                 explicit_null_label_policy_sub_tlv.append(
#                     tunnel_tlv.explicit_null_label_policy_sub_tlv
#                 )
#
#         active_list = self._process_nodes(remote_endpoint_sub_tlv)
#         if active_list != []:
#             self.configure_value(
#                 tunnel_xpath, "enRemoteEndPointTLV", active_list
#             )
#         if any(active_list):
#             self._configure_attributes(
#                 ConfigureBgp._REMOTE_ENDPOINT_SUB_TLV,
#                 remote_endpoint_sub_tlv,
#                 tunnel_xpath,
#             )
#
#         active_list = self._process_nodes(preference_sub_tlv)
#         if active_list != []:
#             self.configure_value(tunnel_xpath, "enPrefTLV", active_list)
#         if any(active_list):
#             self._configure_attributes(
#                 ConfigureBgp._PREFERENCE_SUB_TLV,
#                 preference_sub_tlv,
#                 tunnel_xpath,
#             )
#
#         active_list = self._process_nodes(binding_sub_tlv)
#         if active_list != []:
#             self.configure_value(tunnel_xpath, "enBindingTLV", active_list)
#         if any(active_list):
#             self._configure_attributes(
#                 ConfigureBgp._BINDING_SUB_TLV, binding_sub_tlv, tunnel_xpath
#             )
#
#         active_list = self._process_nodes(explicit_null_label_policy_sub_tlv)
#         if active_list != []:
#             self.configure_value(tunnel_xpath, "enENLPTLV", active_list)
#         if any(active_list):
#             self._configure_attributes(
#                 ConfigureBgp._ENLP_SUB_TLV,
#                 explicit_null_label_policy_sub_tlv,
#                 tunnel_xpath,
#             )
#         self._configure_tlv_segment(ixn_tunnel, tunnel_tlvs)
#
#     def _configure_tlv_segment(self, ixn_tunnel, tunnel_tlvs):
#         nodes_info = self._get_symmetric_nodes(tunnel_tlvs, "segment_lists")
#         if (
#             re.search(
#                 "bgpSRTEPoliciesTunnelEncapsulationListV4", ixn_tunnel.href
#             )
#             is not None
#         ):
#             ixn_tunnel.NumberOfSegmentListV4 = nodes_info.max_len
#             ixn_segment_list = ixn_tunnel.BgpSRTEPoliciesSegmentListV4
#         else:
#             ixn_tunnel.NumberOfSegmentListV6 = nodes_info.max_len
#             ixn_segment_list = ixn_tunnel.BgpSRTEPoliciesSegmentListV6
#         segment_list_xpath = self.get_xpath(ixn_segment_list.href)
#         self.configure_value(
#             segment_list_xpath, "active", nodes_info.active_list
#         )
#         segment_list = nodes_info.symmetric_nodes
#         if any(nodes_info.active_list):
#             self._configure_attributes(
#                 ConfigureBgp._POLICIES_SEGMENT_LIST,
#                 segment_list,
#                 segment_list_xpath,
#             )
#             self.configure_value(
#                 segment_list_xpath, "enWeight", [True] * len(segment_list)
#             )
#         nodes_info = self._get_symmetric_nodes(segment_list, "segments")
#         if (
#             re.search("bgpSRTEPoliciesSegmentListV4", ixn_segment_list.href)
#             is not None
#         ):
#             ixn_segment_list.NumberOfSegmentsV4 = nodes_info.max_len
#             ixn_segments = ixn_segment_list.BgpSRTEPoliciesSegmentsCollectionV4
#         else:
#             ixn_segment_list.NumberOfSegmentsV6 = nodes_info.max_len
#             ixn_segments = ixn_segment_list.BgpSRTEPoliciesSegmentsCollectionV6
#         segments_xpath = self.get_xpath(ixn_segments.href)
#         self.configure_value(segments_xpath, "active", nodes_info.active_list)
#         segments = nodes_info.symmetric_nodes
#         if any(nodes_info.active_list):
#             self._configure_attributes(
#                 ConfigureBgp._SEGMENTS, segments, segments_xpath
#             )
#
#     def _process_nodes(self, nodes):
#         active_list = []
#         for index, node in enumerate(nodes):
#             active = False
#             if node is None:
#                 if index == 0:
#                     nodes[0] = next(v for v in nodes if v is not None)
#                 else:
#                     nodes[index] = nodes[index - 1]
#             else:
#                 is_config = False
#                 for name, value in node._properties.items():
#                     if value is not None:
#                         is_config = True
#                         break
#                 if is_config is True:
#                     active = True
#             active_list.append(active)
#         return active_list
#
#     def _configure_attributes(self, mapper, parent_list, xpath):
#         for attribute in mapper:
#             attr_mapper = mapper[attribute]
#             ixn_attribute = attr_mapper["ixn_attr"]
#             default_value = attr_mapper["default"]
#             enum_map = attr_mapper.get("enum_map")
#             default_obj = getattr(self, str(default_value), None)
#             config_values = []
#             for parent in parent_list:
#                 config_value = getattr(parent, attribute, None)
#                 if config_value is not None:
#                     if enum_map is None:
#                         config_values.append(str(config_value))
#                     else:
#                         if str(config_value) not in enum_map.keys():
#                             raise Exception(
#                                 "{0} must configure with enum {1}".format(
#                                     attribute, enum_map.keys()
#                                 )
#                             )
#                         config_values.append(enum_map[str(config_value)])
#                 elif default_obj is not None:
#                     config_values.append(default_obj())
#                 else:
#                     config_values.append(default_value)
#             self.configure_value(xpath, ixn_attribute, config_values)
#
#     def stop_topology(self):
#         glob_topo = self._api._globals.Topology.refresh()
#         if glob_topo.Status == "started":
#             self._api._ixnetwork.StopAllProtocols("sync")
#
#
# class RouteAddresses(object):
#
#     _IPv4 = "ipv4"
#     _IPv6 = "ipv6"
#
#     def __init__(self):
#         self._address = []
#         self._count = []
#         self._prefix = []
#         self._step = []
#         self._ip_type = None
#
#     def _comp_value(self, values):
#         com_values = []
#         idx = 0
#         while idx < len(values[0]):
#             for value in values:
#                 com_values.append(value[idx])
#             idx += 1
#         return com_values
#
#     @property
#     def address(self):
#         if isinstance(self._address[0], list):
#             return self._comp_value(self._address)
#         return self._address
#
#     @address.setter
#     def address(self, value):
#         self._address.append(value)
#
#     @property
#     def count(self):
#         if isinstance(self._count[0], list):
#             return self._comp_value(self._count)
#         return self._count
#
#     @count.setter
#     def count(self, value):
#         self._count.append(value)
#
#     @property
#     def prefix(self):
#         if isinstance(self._prefix[0], list):
#             return self._comp_value(self._prefix)
#         return self._prefix
#
#     @prefix.setter
#     def prefix(self, value):
#         self._prefix.append(value)
#
#     @property
#     def step(self):
#         if isinstance(self._step[0], list):
#             return self._comp_value(self._step)
#         return self._step
#
#     @step.setter
#     def step(self, value):
#         self._step.append(value)
#
#     def _get_ip_type(self, addresses):
#         class_name = addresses[0].__class__.__name__
#         if re.search("v4", class_name) is not None:
#             return RouteAddresses._IPv4
#         else:
#             return RouteAddresses._IPv6
#
#     def _address_to_int(self, addr):
#         if self._ip_type == RouteAddresses._IPv4:
#             return struct.unpack("!I", socket.inet_aton(addr))[0]
#         else:
#             hi, lo = struct.unpack(
#                 "!QQ", socket.inet_pton(socket.AF_INET6, addr)
#             )
#             return (hi << 64) | lo



from snappi_ixnetwork.device.base import Base
from snappi_ixnetwork.logger import get_ixnet_logger
from snappi_ixnetwork.device.bgpevpn import BgpEvpn


class Bgp(Base):
    _BGP = {
        "peer_address": "dutIp",
        "as_type": {
            "ixn_attr": "type",
            "enum_map": {"ibgp": "internal", "ebgp": "external"},
        },
    }

    _ADVANCED = {
        "hold_time_interval": "holdTimer",
        "keep_alive_interval": "keepaliveTimer",
        "update_interval": "updateInterval",
        "time_to_live": "ttl",
        "md5_key": "md5Key",
    }

    _CAPABILITY = {
        "ipv4_unicast": "capabilityIpV4Unicast",
        "ipv4_multicast": "capabilityIpV4Multicast",
        "ipv6_unicast": "capabilityIpV6Unicast",
        "ipv6_multicast": "capabilityIpV6Multicast",
        "vpls": "capabilityVpls",
        "route_refresh": "capabilityRouteRefresh",
        "route_constraint": "capabilityRouteConstraint",
        "ink_state_non_vpn": "capabilityLinkStateNonVpn",
        "link_state_vpn": "capabilityLinkStateVpn",
        "evpn": "evpn",
        "ipv4_multicast_vpn": "capabilityIpV4MulticastVpn",
        "ipv4_mpls_vpn": "capabilityIpV4MplsVpn",
        "ipv4_mdt": "capabilityIpV4Mdt",
        "ipv4_multicast_mpls_vpn": "ipv4MulticastBgpMplsVpn",
        "ipv4_unicast_flow_spec": "capabilityipv4UnicastFlowSpec",
        "ipv4_sr_te_policy": "capabilitySRTEPoliciesV4",
        "ipv4_unicast_add_path": "capabilityIpv4UnicastAddPath",
        "ipv6_multicast_vpn": "capabilityIpV6MulticastVpn",
        "ipv6_mpls_vpn": "capabilityIpV6MplsVpn",
        "ipv6_multicast_mpls_vpn": "ipv6MulticastBgpMplsVpn",
        "ipv6_unicast_flow_spec": "capabilityipv6UnicastFlowSpec",
        "ipv6_sr_te_policy": "capabilitySRTEPoliciesV6",
        "ipv6_unicast_add_path": "capabilityIpv6UnicastAddPath",
    }

    _CAPABILITY_IPv6 = {
        "extended_next_hop_encoding": "capabilityNHEncodingCapabilities",
        # "ipv6_mdt": "",
    }

    _IP_POOL = {
        "address": "networkAddress",
        "prefix": "prefixLength",
        "count": "numberOfAddressesAsy",
        "step": "prefixAddrStep",
    }

    _ROUTE = {
        "next_hop_mode": {
            "ixn_attr": "nextHopType",
            "enum_map": {"local_ip": "sameaslocalip", "manual": "manually"},
        },
        "next_hop_address_type": "nextHopIPType",
        "next_hop_ipv4_address": "ipv4NextHop",
        "next_hop_ipv6_address": "ipv6NextHop",
    }

    _COMMUNITY = {
        "type": {
            "ixn_attr": "type",
            "enum_map": {
                "manual_as_number": "manual",
                "no_export": "noexport",
                "no_advertised": "noadvertised",
                "no_export_subconfed": "noexport_subconfed",
                "llgr_stale": "llgr_stale",
                "no_llgr": "no_llgr",
            },
        },
        "as_number": "asNumber",
        "as_custom": "lastTwoOctets",
    }

    _BGP_AS_MODE = {
        "do_not_include_local_as": "dontincludelocalas",
        "include_as_seq": "includelocalasasasseq",
        "include_as_set": "includelocalasasasset",
        "include_as_confed_seq": "includelocalasasasseqconfederation",
        "include_as_confed_set": "includelocalasasassetconfederation",
        "prepend_to_first_segment": "prependlocalastofirstsegment",
    }

    _BGP_SEG_TYPE = {
        "as_seq": "asseq",
        "as_set": "asset",
        "as_confed_seq": "asseqconfederation",
        "as_confed_set": "assetconfederation",
    }

    def __init__(self, ngpf):
        super(Bgp, self).__init__()
        self._ngpf = ngpf
        self.logger = get_ixnet_logger(__name__)
        self._bgp_evpn = BgpEvpn(ngpf)
        self._router_id = None

    def config(self, device):
        self.logger.debug("Configuring BGP")
        bgp = device.get("bgp")
        if bgp is None:
            return
        self._router_id = bgp.get("router_id")
        self._config_ipv4_interfaces(bgp)
        self._config_ipv6_interfaces(bgp)

    def _get_interface_info(self):
        ip_types = ["ipv4", "ipv6"]
        same_dg_ips = []
        invalid_ips = []
        ethernets = self._ngpf.working_dg.get("ethernet")
        if ethernets is None:
            return same_dg_ips, invalid_ips
        for ethernet in ethernets:
            for ip_type in ip_types:
                ips = ethernet.get(ip_type)
                if ips is not None:
                    ip_names = [ip.get("name").value for ip in ips]
                    same_dg_ips.extend(ip_names)
                    if len(ips) > 1:
                        invalid_ips.extend(ip_names)
        return same_dg_ips, invalid_ips

    def _is_valid(self, ip_name):
        is_invalid = True
        same_dg_ips, invalid_ips = self._get_interface_info()
        self.logger.debug(
            "Validating %s against interface same_dg_ips : %s invalid_ips %s"
            % (ip_name, same_dg_ips, invalid_ips)
        )
        if ip_name in invalid_ips:
            self._ngpf.api.add_error(
                "Multiple IP {name} on top of name Ethernet".format(
                    name=ip_name
                )
            )
            is_invalid = False
        if len(same_dg_ips) > 0 and ip_name not in same_dg_ips:
            self._ngpf.api.add_error(
                "BGP should not configured on top of different device"
            )
            is_invalid = False
        return is_invalid

    def _config_ipv4_interfaces(self, bgp):
        self.logger.debug("Configuring BGP IPv4 interfaces")
        ipv4_interfaces = bgp.get("ipv4_interfaces")
        if ipv4_interfaces is None:
            return
        for ipv4_interface in ipv4_interfaces:
            ipv4_name = ipv4_interface.get("ipv4_name")
            self._ngpf.working_dg = self._ngpf.api.ixn_objects.get_working_dg(
                ipv4_name
            )
            if not self._is_valid(ipv4_name):
                continue
            ixn_ipv4 = self._ngpf.api.ixn_objects.get_object(ipv4_name)
            self._config_bgpv4(ipv4_interface.get("peers"), ixn_ipv4)

    def _config_ipv6_interfaces(self, bgp):
        self.logger.debug("Configuring BGP IPv6 interfaces")
        ipv6_interfaces = bgp.get("ipv6_interfaces")
        if ipv6_interfaces is None:
            return
        for ipv6_interface in ipv6_interfaces:
            ipv6_name = ipv6_interface.get("ipv6_name")
            self._ngpf.working_dg = self._ngpf.api.ixn_objects.get_working_dg(
                ipv6_name
            )
            if not self._is_valid(ipv6_name):
                continue
            ixn_ipv6 = self._ngpf.api.ixn_objects.get_object(ipv6_name)
            self._config_bgpv6(ipv6_interface.get("peers"), ixn_ipv6)

    def _config_as_number(self, bgp_peer, ixn_bgp):
        as_number_width = bgp_peer.get("as_number_width")
        as_number = bgp_peer.get("as_number")
        if as_number_width == "two":
            ixn_bgp["localAs2Bytes"] = self.multivalue(as_number)
        else:
            ixn_bgp["enable4ByteAs"] = self.multivalue(True)
            ixn_bgp["localAs4Bytes"] = self.multivalue(as_number)

    def _config_bgpv4(self, bgp_peers, ixn_ipv4):
        if bgp_peers is None:
            return
        self.logger.debug("Configuring BGPv4 Peer")
        for bgp_peer in bgp_peers:
            ixn_bgpv4 = self.create_node_elemet(
                ixn_ipv4, "bgpIpv4Peer", bgp_peer.get("name")
            )
            self._ngpf.set_device_info(bgp_peer, ixn_bgpv4)
            self.configure_multivalues(bgp_peer, ixn_bgpv4, Bgp._BGP)
            self._config_as_number(bgp_peer, ixn_bgpv4)
            advanced = bgp_peer.get("advanced")
            if advanced is not None:
                self.configure_multivalues(advanced, ixn_bgpv4, Bgp._ADVANCED)
            capability = bgp_peer.get("capability")
            if capability is not None:
                self.configure_multivalues(
                    capability, ixn_bgpv4, Bgp._CAPABILITY
                )
            self._bgp_route_builder(bgp_peer, ixn_bgpv4)
            self._bgp_evpn.config(bgp_peer, ixn_bgpv4)

    def _config_bgpv6(self, bgp_peers, ixn_ipv6):
        self.logger.debug("Configuring BGPv6 Peer")
        if bgp_peers is None:
            return
        for bgp_peer in bgp_peers:
            ixn_bgpv6 = self.create_node_elemet(
                ixn_ipv6, "bgpIpv6Peer", bgp_peer.get("name")
            )
            self._ngpf.set_device_info(bgp_peer, ixn_bgpv6)
            self.configure_multivalues(bgp_peer, ixn_bgpv6, Bgp._BGP)
            self._config_as_number(bgp_peer, ixn_bgpv6)
            advanced = bgp_peer.get("advanced")
            if advanced is not None:
                self.configure_multivalues(advanced, ixn_bgpv6, Bgp._ADVANCED)
            capability = bgp_peer.get("capability")
            if capability is not None:
                self.configure_multivalues(
                    capability, ixn_bgpv6, Bgp._CAPABILITY
                )
                self.configure_multivalues(
                    capability, ixn_bgpv6, Bgp._CAPABILITY_IPv6
                )
            self._bgp_route_builder(bgp_peer, ixn_bgpv6)
            self._bgp_evpn.config(bgp_peer, ixn_bgpv6)

    def _bgp_route_builder(self, bgp_peer, ixn_bgp):
        v4_routes = bgp_peer.get("v4_routes")
        if v4_routes is not None:
            self._configure_bgpv4_route(v4_routes, ixn_bgp)
        v6_routes = bgp_peer.get("v6_routes")
        if v6_routes is not None:
            self._configure_bgpv6_route(v6_routes, ixn_bgp)
        self._ngpf.compactor.compact(self._ngpf.working_dg.get("networkGroup"))

    def _configure_bgpv4_route(self, v4_routes, ixn_bgp):
        if v4_routes is None:
            return
        self.logger.debug("Configuring BGPv4 Route")
        for route in v4_routes:
            addresses = route.get("addresses")
            for addresse in addresses:
                ixn_ng = self.create_node_elemet(
                    self._ngpf.working_dg, "networkGroup", route.get("name")
                )
                ixn_ng["multiplier"] = 1
                ixn_ip_pool = self.create_node_elemet(
                    ixn_ng, "ipv4PrefixPools", route.get("name")
                )
                ixn_connector = self.create_property(ixn_ip_pool, "connector")
                ixn_connector["connectedTo"] = self.post_calculated(
                    "connectedTo", ref_ixnobj=ixn_bgp
                )
                self.configure_multivalues(addresse, ixn_ip_pool, Bgp._IP_POOL)
                ixn_route = self.create_node_elemet(
                    ixn_ip_pool, "bgpIPRouteProperty", route.get("name")
                )
                self._ngpf.set_device_info(route, ixn_ip_pool)
                self._configure_route(route, ixn_route)

    def _configure_bgpv6_route(self, v6_routes, ixn_bgp):
        if v6_routes is None:
            return
        self.logger.debug("Configuring BGPv6 Route")
        for route in v6_routes:
            addresses = route.get("addresses")
            for addresse in addresses:
                ixn_ng = self.create_node_elemet(
                    self._ngpf.working_dg, "networkGroup", route.get("name")
                )
                ixn_ng["multiplier"] = 1
                ixn_ip_pool = self.create_node_elemet(
                    ixn_ng, "ipv6PrefixPools", route.get("name")
                )
                ixn_connector = self.create_property(ixn_ip_pool, "connector")
                ixn_connector["connectedTo"] = self.post_calculated(
                    "connectedTo", ref_ixnobj=ixn_bgp
                )
                self.configure_multivalues(addresse, ixn_ip_pool, Bgp._IP_POOL)
                ixn_route = self.create_node_elemet(
                    ixn_ip_pool, "bgpV6IPRouteProperty", route.get("name")
                )
                self._ngpf.set_device_info(route, ixn_ip_pool)
                self._configure_route(route, ixn_route)

    def _configure_route(self, route, ixn_route):
        self._ngpf.set_ixn_routes(route, ixn_route)
        self.configure_multivalues(route, ixn_route, Bgp._ROUTE)

        advanced = route.get("advanced")
        if advanced is not None:
            self.logger.debug("Configuring BGP route advance")
            multi_exit_discriminator = advanced.get("multi_exit_discriminator")
            if multi_exit_discriminator is not None:
                ixn_route["enableMultiExitDiscriminator"] = self.multivalue(
                    True
                )
                ixn_route["multiExitDiscriminator"] = self.multivalue(
                    multi_exit_discriminator
                )
            ixn_route["origin"] = self.multivalue(advanced.get("origin"))

        communities = route.get("communities")
        if communities is not None and len(communities) > 0:
            self.logger.debug("Configuring BGP route community")
            ixn_route["enableCommunity"] = self.multivalue(True)
            ixn_route["noOfCommunities"] = len(communities)
            for community in communities:
                ixn_community = self.create_node_elemet(
                    ixn_route, "bgpCommunitiesList"
                )
                self.configure_multivalues(
                    community, ixn_community, Bgp._COMMUNITY
                )

        as_path = route.get("as_path")
        if as_path is not None:
            self.logger.debug("Configuring BGP route AS path")
            ixn_route["enableAsPathSegments"] = self.multivalue(True)
            ixn_route["asSetMode"] = self.multivalue(
                as_path.get("as_set_mode"), Bgp._BGP_AS_MODE
            )
            segments = as_path.get("segments")
            ixn_route["noOfASPathSegmentsPerRouteRange"] = len(segments)
            for segment in segments:
                ixn_segment = self.create_node_elemet(
                    ixn_route, "bgpAsPathSegmentList"
                )
                ixn_segment["segmentType"] = self.multivalue(
                    segment.get(type), Bgp._BGP_SEG_TYPE
                )
                as_numbers = segment.get("as_numbers")
                ixn_segment["numberOfAsNumberInSegment"] = len(as_numbers)
                for as_number in as_numbers:
                    ixn_as_number = self.create_node_elemet(
                        ixn_segment, "bgpAsNumberList"
                    )
                    ixn_as_number["asNumber"] = self.multivalue(as_number)



import time
from collections import OrderedDict

timer_data = OrderedDict()


class Timer:
    def __init__(self, api, msg):
        self._api = api
        self._msg = msg

    def __enter__(self):
        """Start a new timer as a context manager"""
        self._start = time.time()
        return self

    def __exit__(self, *exc_info):
        """Stop the context manager timer"""
        time_lapse = time.time() - self._start
        self._api.info(self._msg + " %.3fs" % (time_lapse))
        timer_data[self._msg] = time_lapse



def test_ports(api, utils):
    """Demonstrates adding ports to a configuration and setting the
    configuration on the traffic generator.
    The traffic generator should have no  items configured other than
    the ports in this test.
    """
    tx_port = utils.settings.ports[0]
    rx_port = utils.settings.ports[1]
    config = api.config()
    config.ports.port(name="tx_port", location=tx_port).port(
        name="rx_port", location=rx_port
    ).port(name="port with no location")
    config.options.port_options.location_preemption = True
    api.set_config(config)
    config = api.config()
    api.set_config(config)



def test_device_lag_name(api, utils):
    config = api.config()
    p1 = config.ports.port(name="p1", location=utils.settings.ports[0])[-1]
    config.options.port_options.location_preemption = True
    lag1 = config.lags.lag(name="lag-1")[-1]
    lp1 = lag1.ports.port(port_name=p1.name)[-1]
    lag1.protocol.static.lag_id = 1
    lp1.ethernet.name = "lp1-e"
    lp1.ethernet.mac = "aa:aa:aa:aa:aa:aa"
    d1 = config.devices.device(name="d1")[-1]
    eth1 = d1.ethernets.ethernet()[-1]
    eth1.connection.lag_name = lag1.name
    eth1.name = "eth1"
    eth1.mac = "00:01:00:00:00:02"
    api.set_config(config)
    assert (api._ixnetwork.Lag.find()[0].Name) == lag1.name
    

def test_port_name(api, utils):
    config = api.config()
    p1 = config.ports.port(name="p1", location=utils.settings.ports[0])[-1]
    config.options.port_options.location_preemption = True
    d1 = config.devices.device(name="d1")[-1]
    eth1 = d1.ethernets.ethernet()[-1]
    eth1.connection.port_name = p1.name
    eth1.name = "eth1"
    eth1.mac = "00:01:00:00:00:01"
    api.set_config(config)
    validate_config(api, p1.name)


def test_connection_portname(api, utils):
    config = api.config()
    p1 = config.ports.port(name="p1", location=utils.settings.ports[0])[-1]
    config.options.port_options.location_preemption = True
    d1 = config.devices.device(name="d1")[-1]
    eth1 = d1.ethernets.ethernet()[-1]
    eth1.connection.port_name = p1.name
    eth1.name = "eth1"
    eth1.mac = "00:01:00:00:00:01"
    api.set_config(config)
    validate_config(api, p1.name)
    api._ixnetwork.NewConfig()


def test_device_connection(api, utils):
    """
    Test when both port_name and connection.port_name is set in the config
    """
    config = api.config()
    p1, p2 = config.ports.port(
        name="p1", location=utils.settings.ports[0]
    ).port(name="p2", location=utils.settings.ports[1])
    config.options.port_options.location_preemption = True
    d1 = config.devices.device(name="d1")[-1]
    eth1 = d1.ethernets.add()
    eth1.connection.port_name = p2.name
    eth1.connection.port_name = p1.name
    eth1.name = "eth1"
    eth1.mac = "00:01:00:00:00:01"
    try:
        api.set_config(config)
    except Exception as err:
        assert err.args[0] == 500


def test_device_without_port_name(api, utils):
    config = api.config()
    p1 = config.ports.port(name="p1", location=utils.settings.ports[0])[-1]
    config.options.port_options.location_preemption = True
    d1 = config.devices.device(name="d1")[-1]
    eth1 = d1.ethernets.ethernet()[-1]
    eth1.name = "eth1"
    eth1.mac = "00:01:00:00:00:01"
    try:
        api.set_config(config)
    except Exception as err:
        assert (
            str(err).split("\n ")[-1]
        ) == "port_name is not passed for the device d1"


def validate_config(api, port_name):
    assert (api._ixnetwork.Vport.find()[0].Name) == port_name



import pytest


def test_static_lag(api, utils):
    """Demonstrates the following:
    1) Creating a lag comprised of multiple ports
    2) Creating emulated devices over the lag
    3) Creating traffic over the emulated devices that will transmit
    traffic to a single rx port.

        TX LAG              DUT             RX
        ------+         +---------+
        port 1|         |
        ..    | ------> |
        port n|         |
        ------+
    """
    config = api.config()
    p1, p2, p3, p4 = (
        config.ports.port(name="txp1", location=utils.settings.ports[0])
        .port(name="txp2", location=utils.settings.ports[1])
        .port(name="rxp1", location=utils.settings.ports[2])
        .port(name="rxp2", location=utils.settings.ports[3])
    )

    config.layer1.layer1(
        name="layer1",
        port_names=[p.name for p in config.ports],
        speed=utils.settings.speed,
        media=utils.settings.media,
    )

    lag1, lag2 = config.lags.lag(name="lag1").lag(name="lag2")
    lp1, lp2 = lag1.ports.port(port_name=p1.name).port(port_name=p2.name)
    lp3, lp4 = lag2.ports.port(port_name=p3.name).port(port_name=p4.name)

    lag1.protocol.static.lag_id = 1
    lag2.protocol.static.lag_id = 2

    lp1.ethernet.name, lp2.ethernet.name = "eth1", "eth2"
    lp3.ethernet.name, lp4.ethernet.name = "eth3", "eth4"

    lp1.ethernet.mac = "00:11:02:00:00:01"
    lp2.ethernet.mac = "00:22:02:00:00:01"
    lp3.ethernet.mac = "00:33:02:00:00:01"
    lp4.ethernet.mac = "00:44:02:00:00:01"

    lp1.ethernet.vlans.vlan(priority=1, name="vlan1", id=1)[-1]
    lp2.ethernet.vlans.vlan(priority=1, name="vlan2", id=1)[-1]
    lp3.ethernet.vlans.vlan(priority=1, name="vlan3", id=1)[-1]
    lp4.ethernet.vlans.vlan(priority=1, name="vlan4", id=1)[-1]

    packets = 2000
    f1_size = 74
    f2_size = 1500
    d1, d2 = config.devices.device(name="device1").device(name="device2")
    eth1, eth2 = d1.ethernets.add(), d2.ethernets.add()
    eth1.connection.port_name, eth2.connection.port_name = lag1.name, lag2.name
    eth1.name, eth2.name = "d_eth1", "d_eth2"
    eth1.mac, eth2.mac = "00:00:00:00:00:11", "00:00:00:00:00:22"
    ip1, ip2 = eth1.ipv4_addresses.add(), eth2.ipv4_addresses.add()
    ip1.name, ip2.name = "ip1", "ip2"
    ip1.address = "10.1.1.1"
    ip1.gateway = "10.1.1.2"
    ip2.address = "10.1.1.2"
    ip2.gateway = "10.1.1.1"
    f1, f2 = config.flows.flow(name="f1").flow(name="f2")
    f1.tx_rx.port.tx_name, f1.tx_rx.port.rx_name = p1.name, p2.name
    f2.tx_rx.port.tx_name, f2.tx_rx.port.rx_name = p3.name, p4.name
    config.options.port_options.location_preemption = True
    f1.duration.fixed_packets.packets = packets
    f2.duration.fixed_packets.packets = packets
    f1.size.fixed = f1_size
    f2.size.fixed = f2_size
    f1.rate.percentage = 10
    f2.rate.percentage = 10

    f1.metrics.enable = True
    f1.metrics.loss = True

    f2.metrics.enable = True
    f2.metrics.loss = True

    utils.start_traffic(api, config, start_capture=False)
    utils.wait_for(lambda: utils.is_traffic_stopped(api), "traffic to stop")

    utils.wait_for(
        lambda: utils.is_stats_accumulated(api, packets * 2),
        "stats to be accumulated",
    )

    utils.wait_for(
        lambda: results_ok(api, utils, f1_size, f2_size, packets),
        "stats to be as expected",
        timeout_seconds=30,
    )


def test_lacp_lag(api, utils):
    """Demonstrates the following:
    1) Creating a lag comprised of multiple ports
    2) Creating emulated devices over the lag
    3) Creating traffic over the emulated devices that will transmit
    traffic to a single rx port.

        TX LAG              DUT             RX
        ------+         +---------+
        port 1|         |
        ..    | ------> |
        port n|         |
        ------+
    """
    LAG1_ATTR = {
        "ActorSystemId": ["00 11 03 00 00 03", "00 11 03 00 00 03"],
        "ActorSystemPriority": ["1", "1"],
        "ActorKey": ["10", "10"],
        "ActorPortNumber": ["30", "40"],
        "ActorPortPriority": ["100", "101"],
        "LacpActivity": ["active", "active"],
        "LacpduPeriodicTimeInterval": ["5", "6"],
        "LacpduTimeout": ["12", "13"],
    }

    LAG2_ATTR = {
        "ActorSystemId": ["00 22 03 00 00 03", "00 22 03 00 00 03"],
        "ActorSystemPriority": ["2", "2"],
        "ActorKey": ["20", "20"],
        "ActorPortNumber": ["50", "60"],
        "ActorPortPriority": ["200", "201"],
        "LacpActivity": ["active", "active"],
        "LacpduPeriodicTimeInterval": ["7", "8"],
        "LacpduTimeout": ["14", "15"],
    }
    config = api.config()
    p1, p2, p3, p4 = (
        config.ports.port(name="txp1", location=utils.settings.ports[0])
        .port(name="txp2", location=utils.settings.ports[2])
        .port(name="rxp1", location=utils.settings.ports[1])
        .port(name="rxp2", location=utils.settings.ports[3])
    )

    config.layer1.layer1(
        name="layer1",
        port_names=[p.name for p in config.ports],
        speed=utils.settings.speed,
        media=utils.settings.media,
    )

    lag1, lag2 = config.lags.lag(name="lag1").lag(name="lag2")
    l1_p1, l1_p2 = lag1.ports.port(port_name=p1.name).port(port_name=p2.name)
    l2_p1, l2_p2 = lag2.ports.port(port_name=p3.name).port(port_name=p4.name)
    config.options.port_options.location_preemption = True

    lag1.protocol.lacp.actor_system_id = "00:11:03:00:00:03"
    lag1.protocol.lacp.actor_system_priority = int(
        LAG1_ATTR["ActorSystemPriority"][0]
    )
    lag1.protocol.lacp.actor_key = int(LAG1_ATTR["ActorKey"][0])

    lag2.protocol.lacp.actor_system_id = "00:22:03:00:00:03"
    lag2.protocol.lacp.actor_system_priority = int(
        LAG2_ATTR["ActorSystemPriority"][0]
    )
    lag2.protocol.lacp.actor_key = int(LAG2_ATTR["ActorKey"][0])

    l1_p1.lacp.actor_port_number = int(LAG1_ATTR["ActorPortNumber"][0])
    l1_p2.lacp.actor_port_number = int(LAG1_ATTR["ActorPortNumber"][1])
    l2_p1.lacp.actor_port_number = int(LAG2_ATTR["ActorPortNumber"][0])
    l2_p2.lacp.actor_port_number = int(LAG2_ATTR["ActorPortNumber"][1])

    l1_p1.lacp.actor_port_priority = int(LAG1_ATTR["ActorPortPriority"][0])
    l1_p2.lacp.actor_port_priority = int(LAG1_ATTR["ActorPortPriority"][1])
    l2_p1.lacp.actor_port_priority = int(LAG2_ATTR["ActorPortPriority"][0])
    l2_p2.lacp.actor_port_priority = int(LAG2_ATTR["ActorPortPriority"][1])

    l1_p1.lacp.actor_activity = LAG1_ATTR["LacpActivity"][0]
    l1_p2.lacp.actor_activity = LAG1_ATTR["LacpActivity"][1]
    l2_p1.lacp.actor_activity = LAG2_ATTR["LacpActivity"][0]
    l2_p2.lacp.actor_activity = LAG2_ATTR["LacpActivity"][1]

    l1_p1.lacp.lacpdu_periodic_time_interval = int(
        LAG1_ATTR["LacpduPeriodicTimeInterval"][0]
    )
    l1_p2.lacp.lacpdu_periodic_time_interval = int(
        LAG1_ATTR["LacpduPeriodicTimeInterval"][1]
    )
    l2_p1.lacp.lacpdu_periodic_time_interval = int(
        LAG2_ATTR["LacpduPeriodicTimeInterval"][0]
    )
    l2_p2.lacp.lacpdu_periodic_time_interval = int(
        LAG2_ATTR["LacpduPeriodicTimeInterval"][1]
    )

    l1_p1.lacp.lacpdu_timeout = int(LAG1_ATTR["LacpduTimeout"][0])
    l1_p2.lacp.lacpdu_timeout = int(LAG1_ATTR["LacpduTimeout"][1])
    l2_p1.lacp.lacpdu_timeout = int(LAG2_ATTR["LacpduTimeout"][0])
    l2_p2.lacp.lacpdu_timeout = int(LAG2_ATTR["LacpduTimeout"][1])

    l1_p1.ethernet.name, l1_p2.ethernet.name = "eth1", "eth2"
    l2_p1.ethernet.name, l2_p2.ethernet.name = "eth3", "eth4"

    l1_p1.ethernet.mac = "00:11:02:00:00:01"
    l1_p2.ethernet.mac = "00:22:02:00:00:01"
    l2_p1.ethernet.mac = "00:33:02:00:00:01"
    l2_p2.ethernet.mac = "00:44:02:00:00:01"

    l1_p1.ethernet.vlans.vlan(priority=1, name="vlan1", id=1)[-1]
    l1_p2.ethernet.vlans.vlan(priority=1, name="vlan2", id=1)[-1]
    l2_p1.ethernet.vlans.vlan(priority=1, name="vlan3", id=1)[-1]
    l2_p2.ethernet.vlans.vlan(priority=1, name="vlan4", id=1)[-1]

    d1, d2 = config.devices.device(name="device1").device(name="device2")
    eth1, eth2 = d1.ethernets.add(), d2.ethernets.add()
    eth1.connection.port_name, eth2.connection.port_name = lag1.name, lag2.name
    eth1.name, eth2.name = "d_eth1", "d_eth2"
    eth1.mac, eth2.mac = "00:00:00:00:00:11", "00:00:00:00:00:22"
    ip1, ip2 = eth1.ipv4_addresses.add(), eth2.ipv4_addresses.add()
    ip1.name, ip2.name = "ip1", "ip2"
    ip1.address = "10.1.1.1"
    ip1.gateway = "10.1.1.2"
    ip2.address = "10.1.1.2"
    ip2.gateway = "10.1.1.1"

    api.set_config(config)

    utils.wait_for(
        lambda: lacp_pdu_status_ok(api, "up"), "port state as expected"
    )

    cs = api.control_state()
    cs.protocol.lacp.member_ports.lag_member_names = ["rxp2"]
    cs.protocol.lacp.member_ports.state = cs.protocol.lacp.member_ports.DOWN
    api.set_control_state(cs)

    utils.wait_for(
        lambda: lacp_pdu_status_ok(api, "down"), "port state as expected"
    )

    cs = api.control_state()
    cs.protocol.lacp.member_ports.lag_member_names = ["rxp2"]
    cs.protocol.lacp.member_ports.state = cs.protocol.lacp.member_ports.UP
    api.set_control_state(cs)

    utils.wait_for(
        lambda: lacp_pdu_status_ok(api, "up"), "port state as expected"
    )

    validate_lacp_config(api, LAG1_ATTR, LAG2_ATTR)


def lacp_pdu_status_ok(api, state):
    return (
        api._ixnetwork.Lag.find()
        .ProtocolStack.find()
        .Ethernet.find()
        .Lagportlacp.find()
        .SessionStatus
    )[1] == state


def validate_lacp_config(api, LAG1_ATTR, LAG2_ATTR):
    lags = api._ixnetwork.Lag.find()

    for lag in lags:
        assert lag.LagMode.LagProtocol == "lacp"

    lag1 = lags[0].ProtocolStack.find().Ethernet.find().Lagportlacp.find()
    for attr in LAG1_ATTR:
        assert getattr(lag1, attr).Values == LAG1_ATTR[attr]

    lag2 = lags[1].ProtocolStack.find().Ethernet.find().Lagportlacp.find()
    for attr in LAG2_ATTR:
        assert getattr(lag2, attr).Values == LAG2_ATTR[attr]


def test_static_and_lacp_lag(api, utils):
    LACP_ATTR = {
        "ActorSystemId": ["00 22 03 00 00 03", "00 22 03 00 00 03"],
        "ActorSystemPriority": ["1", "1"],
        "ActorKey": ["10", "10"],
        "ActorPortNumber": ["30", "40"],
        "ActorPortPriority": ["100", "101"],
        "LacpActivity": ["active", "passive"],
        "LacpduPeriodicTimeInterval": ["5", "6"],
        "LacpduTimeout": ["12", "13"],
    }

    config = api.config()
    p1, p2, p3, p4 = (
        config.ports.port(name="txp1", location=utils.settings.ports[0])
        .port(name="txp2", location=utils.settings.ports[2])
        .port(name="rxp1", location=utils.settings.ports[1])
        .port(name="rxp2", location=utils.settings.ports[3])
    )

    config.layer1.layer1(
        name="layer1",
        port_names=[p.name for p in config.ports],
        speed=utils.settings.speed,
        media=utils.settings.media,
    )

    lag1, lag2 = config.lags.lag(name="lag1").lag(name="lag2")
    l1_p1, l1_p2 = lag1.ports.port(port_name=p1.name).port(port_name=p2.name)
    l2_p1, l2_p2 = lag2.ports.port(port_name=p3.name).port(port_name=p4.name)
    config.options.port_options.location_preemption = True

    lag1.protocol.static.lag_id = 5

    lag2.protocol.lacp.actor_system_id = "00:22:03:00:00:03"
    lag2.protocol.lacp.actor_system_priority = int(
        LACP_ATTR["ActorSystemPriority"][0]
    )
    lag2.protocol.lacp.actor_key = int(LACP_ATTR["ActorKey"][0])

    l2_p1.lacp.actor_port_number = int(LACP_ATTR["ActorPortNumber"][0])
    l2_p2.lacp.actor_port_number = int(LACP_ATTR["ActorPortNumber"][1])

    l2_p1.lacp.actor_port_priority = int(LACP_ATTR["ActorPortPriority"][0])
    l2_p2.lacp.actor_port_priority = int(LACP_ATTR["ActorPortPriority"][1])

    l2_p1.lacp.actor_activity = LACP_ATTR["LacpActivity"][0]
    l2_p2.lacp.actor_activity = LACP_ATTR["LacpActivity"][1]

    l2_p1.lacp.lacpdu_periodic_time_interval = int(
        LACP_ATTR["LacpduPeriodicTimeInterval"][0]
    )
    l2_p2.lacp.lacpdu_periodic_time_interval = int(
        LACP_ATTR["LacpduPeriodicTimeInterval"][1]
    )

    l2_p1.lacp.lacpdu_timeout = int(LACP_ATTR["LacpduTimeout"][0])
    l2_p2.lacp.lacpdu_timeout = int(LACP_ATTR["LacpduTimeout"][1])

    l1_p1.ethernet.name, l1_p2.ethernet.name = "eth1", "eth2"
    l2_p1.ethernet.name, l2_p2.ethernet.name = "eth3", "eth4"

    l1_p1.ethernet.mac = "00:11:02:00:00:01"
    l1_p2.ethernet.mac = "00:22:02:00:00:01"
    l2_p1.ethernet.mac = "00:33:02:00:00:01"
    l2_p2.ethernet.mac = "00:44:02:00:00:01"

    l1_p1.ethernet.vlans.vlan(priority=1, name="vlan1", id=1)[-1]
    l1_p2.ethernet.vlans.vlan(priority=1, name="vlan2", id=1)[-1]
    l2_p1.ethernet.vlans.vlan(priority=1, name="vlan3", id=1)[-1]
    l2_p2.ethernet.vlans.vlan(priority=1, name="vlan4", id=1)[-1]

    packets = 2000
    f1_size = 74
    f2_size = 1500
    d1, d2 = config.devices.device(name="device1").device(name="device2")
    eth1, eth2 = d1.ethernets.add(), d2.ethernets.add()
    eth1.connection.port_name, eth2.connection.port_name = lag1.name, lag2.name
    eth1.name, eth2.name = "d_eth1", "d_eth2"
    eth1.mac, eth2.mac = "00:00:00:00:00:11", "00:00:00:00:00:22"
    ip1, ip2 = eth1.ipv4_addresses.add(), eth2.ipv4_addresses.add()
    ip1.name, ip2.name = "ip1", "ip2"
    ip1.address = "10.1.1.1"
    ip1.gateway = "10.1.1.2"
    ip2.address = "10.1.1.2"
    ip2.gateway = "10.1.1.1"
    f1, f2 = config.flows.flow(name="f1").flow(name="f2")
    f1.tx_rx.port.tx_name, f1.tx_rx.port.rx_name = p1.name, p2.name
    f2.tx_rx.port.tx_name, f2.tx_rx.port.rx_name = p3.name, p4.name
    f1.duration.fixed_packets.packets = packets
    f2.duration.fixed_packets.packets = packets
    f1.size.fixed = f1_size
    f2.size.fixed = f2_size
    f1.rate.percentage = 10
    f2.rate.percentage = 10

    api.set_config(config)
    validate_static_lacp_config(api, LACP_ATTR)


def validate_static_lacp_config(api, LACP_ATTR):
    lags = api._ixnetwork.Lag.find()

    assert (
        lags[0]
        .ProtocolStack.find()
        .Ethernet.find()
        .Lagportstaticlag.find()
        .LagId.Values
    )[0] == "5"
    assert lags[1].LagMode.LagProtocol == "lacp"

    lag_lacp = lags[1].ProtocolStack.find().Ethernet.find().Lagportlacp.find()
    for attr in LACP_ATTR:
        assert getattr(lag_lacp, attr).Values == LACP_ATTR[attr]


def results_ok(api, utils, size1, size2, packets):
    """
    Returns true if stats are as expected, false otherwise.
    """
    port_results, flow_results = utils.get_all_stats(api)
    frames_ok = utils.total_frames_ok(port_results, flow_results, packets * 2)
    bytes_ok = utils.total_bytes_ok(
        port_results, flow_results, packets * size1 + packets * size2
    )
    return frames_ok and bytes_ok


if __name__ == "__main__":
    pytest.main(["-s", __file__])



import json, re

from snappi_ixnetwork.timer import Timer
from snappi_ixnetwork.device.base import Base
from snappi_ixnetwork.device.bgp import Bgp
from snappi_ixnetwork.device.macsec import Macsec
from snappi_ixnetwork.logger import get_ixnet_logger
from snappi_ixnetwork.device.vxlan import VXLAN
from snappi_ixnetwork.device.interface import Ethernet
from snappi_ixnetwork.device.loopbackint import LoopbackInt
from snappi_ixnetwork.device.compactor import Compactor
from snappi_ixnetwork.device.createixnconfig import CreateIxnConfig
from snappi_ixnetwork.device.rocev2 import RoCEv2
from snappi_ixnetwork.device.isis import Isis


class Ngpf(Base):
    _DEVICE_ENCAP_MAP = {
        "Device": "",
        "DeviceEthernet": "ethernetVlan",
        "DeviceIpv4": "ipv4",
        "DeviceIpv6": "ipv6",
        "BgpV4Peer": "ipv4",
        "BgpV6Peer": "ipv6",
        "BgpV4RouteRange": "ipv4",
        "BgpV6RouteRange": "ipv6",
        "DeviceIpv4Loopback": "ipv4",
        "DeviceIpv6Loopback": "ipv6",
        "VxlanV4Tunnel": "ipv4",
        "VxlanV6Tunnel": "ipv6",
        "BgpCMacIpRange": "ethernetVlan",
        "Mka": "ethernetVlan",
        "SecureEntity": "ethernetVlan",
        "Rocev2V4Peer": "ipv4",
        "Rocev2V6Peer": "ipv6",
        "Isis": "ethernetVlan",
        "IsisInterface": "ethernetVlan",
        "IsisV4RouteRange": "ipv4",
        "IsisV6RouteRange": "ipv6",
    }

    _ROUTE_STATE = {"advertise": True, "withdraw": False}

    def __init__(self, ixnetworkapi):
        super(Ngpf, self).__init__()
        self.api = ixnetworkapi
        self._ixn_config = {}
        self.working_dg = None
        self._ixn_topo_objects = {}
        self.ether_v4gateway_map = {}
        self.ether_v6gateway_map = {}
        self.ether_ip_restriction_map = {}
        self.logger = get_ixnet_logger(__name__)
        self._ethernet = Ethernet(self)
        self._bgp = Bgp(self)
        self._macsec = Macsec(self)
        self._vxlan = VXLAN(self)
        self._rocev2 = RoCEv2(self)
        self._isis = Isis(self)
        self._loop_back = LoopbackInt(self)
        self.compactor = Compactor(self.api)
        self._createixnconfig = CreateIxnConfig(self)

    def config(self):
        self._ixn_topo_objects = {}
        self.working_dg = None
        self._ixn_config = dict()
        self.ether_v4gateway_map = {}
        self.ether_v6gateway_map = {}
        self.ether_ip_restriction_map = {}
        self._chain_parent_dgs = []
        self.loopback_parent_dgs = []
        self._ixn_config["xpath"] = "/"
        self._resource_manager = self.api._ixnetwork.ResourceManager
        with Timer(self.api, "Convert device config :"):
            self._configure_topology()
        with Timer(self.api, "Create IxNetwork device config :"):
            self._createixnconfig.create(
                self._ixn_config["topology"], "topology"
            )
            self._createixnconfig.post_calculate()
        with Timer(self.api, "Push IxNetwork device config :"):
            self._pushixnconfig()

    def set_device_info(self, snappi_obj, ixn_obj):
        name = snappi_obj.get("name")
        class_name = snappi_obj.__class__.__name__
        self.logger.debug(
            "set_device_info name %s and class_name %s" % (name, class_name)
        )
        try:
            encap = Ngpf._DEVICE_ENCAP_MAP[class_name]
        except KeyError:
            raise NameError("Mapping is missing for {0}".format(class_name))
        self.api.set_device_encap(name, encap)
        self.api.set_device_encap(self.get_name(self.working_dg), encap)
        self.api.ixn_objects.set(name, ixn_obj)

    def set_ixn_routes(self, snappi_obj, ixn_obj):
        name = snappi_obj.get("name")
        self.logger.debug("set_ixn_routes for %s " % name)
        self.api.ixn_routes.set(name, ixn_obj)

    def _get_topology_name(self, port_name):
        return "Topology %s" % port_name

    def _set_dev_compacted(self, dgs):
        if dgs is None:
            return
        for dg in dgs:
            names = dg.get("name")
            if isinstance(names, list) and len(names) > 1:
                self.api.set_dev_compacted(names[0], names)

    def _configure_topology(self):
        self.stop_topology()
        self.api._remove(self.api._topology, [])
        ixn_topos = self.create_node(self._ixn_config, "topology")
        # Configured all interfaces
        self._configure_device_group(ixn_topos)

        # Configure all MACsec interface before configure protocols
        for device in self.api.snappi_config.devices:
            self._macsec.config(device)

        # We need to configure all interface before configure protocols
        for device in self.api.snappi_config.devices:
            self._bgp.config(device)

        # Configure all RoCEv2 interface before configure protocols
        for device in self.api.snappi_config.devices:
            self._rocev2.config(device)
        
        # Configure all Isis interface before configure protocols
        for device in self.api.snappi_config.devices:
            self._isis.config(device)

        # Compaction will take place in this order
        # Step-1: Compact chain DGs
        for chain_parent_dg in self._chain_parent_dgs:
            self.compactor.compact(chain_parent_dg.get("deviceGroup"))
            self._set_dev_compacted(chain_parent_dg.get("deviceGroup"))

        # Step-2: Compact VXLAN
        source_interfaces = self._vxlan.source_interfaces
        for v4_int in source_interfaces.ipv4:
            self.compactor.compact(v4_int.get("vxlan"))
        for v6_int in source_interfaces.ipv6:
            self.compactor.compact(v6_int.get("vxlanv6"))

        # Step-3: First compact all loopback interfaces
        for ix_parent_dg in self.loopback_parent_dgs:
            self.compactor.compact(ix_parent_dg.get("deviceGroup"))

        # Step-4: Compact root Topology
        for ixn_topo in self._ixn_topo_objects.values():
            self.compactor.compact(ixn_topo.get("deviceGroup"))
            self._set_dev_compacted(ixn_topo.get("deviceGroup"))

        if self.api._port_compaction:
            self.compactor.compact(self._ixn_config.get("topology"), True)

    # internal API
    def _set_ip_restriction(self):
        self.logger.debug("Checking if IPv4/ v6 is restricted when MACsec is present")
        for device in self.api.snappi_config.devices:
            self._macsec.set_ip_restriction(device)

    # external API to be called e.g. from MACsec
    def set_ip_restricted(self, ethernet_name):
        self.ether_ip_restriction_map[ethernet_name] = True
        self.logger.debug("Set IPv4/ v6 restricted for ethernet %s e.g. when MACsec is present" % ethernet_name)

    # external API to be called e.g. from interace
    def is_ip_restricted(self, ethernet_name):
        is_restricted = False
        try:
            is_restricted = self.ether_ip_restriction_map[ethernet_name]
        except KeyError:
            pass
        self.logger.debug("IPv4/ v6 restricted for ethernet %s: %s" % (ethernet_name, "True" if is_restricted == True else "False"))
        return is_restricted

    def _configure_device_group(self, ixn_topos):
        """map ethernet with a ixn deviceGroup with multiplier = 1"""
        port_name = None
        device_chain_dgs = {}
        self._set_ip_restriction()
        for device in self.api.snappi_config.devices:
            chin_dgs = {}
            ethernets = device.get("ethernets")
            if ethernets is None:
                continue
            for ethernet in ethernets:
                if ethernet.get("connection") and ethernet.get("port_name"):
                    raise Exception(
                        "port_name and connection for ethernet configuration cannot be passed together, use either connection or port_name property. \
                            port_name is deprecated and will be removed in future releases."
                    )
                if ethernet.get("connection"):
                    connection_choice = ethernet.get("connection").choice
                    if connection_choice == "port_name":
                        port_name = ethernet.get("connection").port_name
                    elif connection_choice == "lag_name":
                        port_name = ethernet.get("connection").lag_name
                    elif connection_choice == "vxlan_name":
                        port_name = ethernet.get("connection").vxlan_name
                        if port_name in chin_dgs:
                            chin_dgs[port_name].append(ethernet)
                        else:
                            chin_dgs[port_name] = [ethernet]
                        continue
                else:
                    port_name = ethernet.get("port_name")
                if port_name is None:
                    raise Exception(
                        "port_name is not passed for the device {}".format(
                            device.get("name")
                        )
                    )
                if port_name in self._ixn_topo_objects:
                    ixn_topo = self._ixn_topo_objects[port_name]
                else:
                    ixn_topo = self.add_element(ixn_topos)
                    ixn_topo["name"] = self._get_topology_name(port_name)
                    ixn_topo["ports"] = [
                        self.api.ixn_objects.get_xpath(port_name)
                    ]
                    self._ixn_topo_objects[port_name] = ixn_topo
                ixn_dg = self.create_node_elemet(
                    ixn_topo, "deviceGroup", device.get("name")
                )
                ixn_dg["multiplier"] = 1
                self.working_dg = ixn_dg
                self.set_device_info(device, ixn_dg)
                self._ethernet.config(ethernet, ixn_dg)
            device_chain_dgs[device.name] = chin_dgs

        # Create all ethernet before start loopback
        self.loopback_parent_dgs = self._loop_back.config()

        for device in self.api.snappi_config.devices:
            vxlan = device.get("vxlan")
            if vxlan is None:
                continue
            self.working_dg = self.api.ixn_objects.get_working_dg(device.name)
            self._vxlan.config(vxlan)

        # Wait till all primary DG will configure
        for device in self.api.snappi_config.devices:
            chin_dgs = device_chain_dgs.get(device.name)
            if chin_dgs is None:
                continue
            for connected_to, ethernet_list in chin_dgs.items():
                ixn_working_dg = self.api.ixn_objects.get_working_dg(
                    connected_to
                )
                self._chain_parent_dgs.append(ixn_working_dg)
                for ethernet in ethernet_list:
                    ixn_dg = self.create_node_elemet(
                        ixn_working_dg, "deviceGroup", device.get("name")
                    )
                    ixn_dg["multiplier"] = 1
                    self.working_dg = ixn_dg
                    self.set_device_info(device, ixn_dg)
                    self._ethernet.config(ethernet, ixn_dg)

    def _pushixnconfig(self):
        self.logger.debug("pushing ixnet config")
        erros = self.api.get_errors()
        if len(erros) > 0:
            return
        ixn_cnf = json.dumps(self._ixn_config, indent=2)
        errata = self._resource_manager.ImportConfig(ixn_cnf, False)
        for item in errata:
            self.api.warning(item)

    def stop_topology(self):
        glob_topo = self.api._globals.Topology.refresh()
        if glob_topo.Status == "started":
            self.logger.debug("Stopping topology")
            self.api._ixnetwork.StopAllProtocols("sync")

    def set_protocol_state(self, request):
        if request.state is None:
            raise Exception("state is None within set_protocol_state")
        self.logger.debug("Setting protocol with %s" % request.state)
        if request.state == "start":
            if len(self.api._topology.find()) > 0:
                self.api._ixnetwork.StartAllProtocols("sync")
                self.api.check_protocol_statistics()
        else:
            if len(self.api._topology.find()) > 0:
                self.api._ixnetwork.StopAllProtocols("sync")

    def set_route_state(self, payload):
        if payload.state is None:
            return
        names = payload.names
        if len(names) == 0:
            names = self.api.ixn_routes.names
        ixn_obj_idx_list = {}
        names = list(set(names))
        self.logger.debug("set route state for %s" % names)
        for name in names:
            route_info = self.api.ixn_routes.get(name)
            ixn_obj = None
            for obj in ixn_obj_idx_list.keys():
                if obj.xpath == route_info.xpath:
                    ixn_obj = obj
                    break
            if ixn_obj is None:
                ixn_obj_idx_list[route_info] = list(
                    range(
                        route_info.index,
                        route_info.index + route_info.multiplier,
                    )
                )
            else:
                ixn_obj_idx_list[route_info].extend(
                    list(
                        range(
                            route_info.index,
                            route_info.index + route_info.multiplier,
                        )
                    )
                )
        imports = []
        for obj, index_list in ixn_obj_idx_list.items():
            xpath = obj.xpath
            active = "active"
            index_list = list(set(index_list))
            object_info = self.select_properties(xpath, properties=[active])
            values = object_info[active]["values"]
            for idx in index_list:
                values[idx] = Ngpf._ROUTE_STATE[payload.state]
            imports.append(self.configure_value(xpath, active, values))
        self.imports(imports)
        self.api._ixnetwork.Globals.Topology.ApplyOnTheFly()
        return names

    def set_device_state(self, payload):
        lmp_names = payload.member_ports.lag_member_names
        state = payload.member_ports.state
        if lmp_names is None:
            self._lacp_start_stop_pdu(state, 1)
        else:
            for lag, ports in self.api.lag._lag_ports.items():
                for index, port in enumerate(ports):
                    for lmp_n in lmp_names:
                        if port.port_name == lmp_n:
                            self._lacp_start_stop_pdu(state, index + 1, lag)

    def _lacp_start_stop_pdu(self, state, index=1, lag_name=None):
        if lag_name is None:
            lag_port_lacp = (
                self.api._ixnetwork.Lag.find()
                .ProtocolStack.find()
                .Ethernet.find()
                .Lagportlacp.find()
            )
            if state == "up":
                lag_port_lacp.LacpStartPDU()
            elif state == "down":
                lag_port_lacp.LacpStopPDU()
        else:
            lag_port_lacp = (
                self.api._ixnetwork.Lag.find(Name=lag_name)
                .ProtocolStack.find()
                .Ethernet.find()
                .Lagportlacp.find()
            )
            if state == "up":
                lag_port_lacp.LacpStartPDU(SessionIndices=index)
            elif state == "down":
                lag_port_lacp.LacpStopPDU(SessionIndices=index)

    def get_states(self, request):
        self.logger.debug("get_states for %s" % request.choice)
        if request.choice == "ipv4_neighbors":
            ip_objs = (
                self.api._ixnetwork.Topology.find()
                .DeviceGroup.find()
                .Ethernet.find()
                .Ipv4.find()
            )
            resolved_mac_list = self._get_ether_resolved_mac(
                ip_objs,
                self.ether_v4gateway_map,
                request.ipv4_neighbors,
                "ipv4",
            )
        elif request.choice == "ipv6_neighbors":
            ip_objs = (
                self.api._ixnetwork.Topology.find()
                .DeviceGroup.find()
                .Ethernet.find()
                .Ipv6.find()
            )
            resolved_mac_list = self._get_ether_resolved_mac(
                ip_objs,
                self.ether_v6gateway_map,
                request.ipv6_neighbors,
                "ipv6",
            )
        else:
            raise TypeError(
                "get_states only accept ipv4_neighbors or ipv6_neighbors"
            )

        return {"choice": request.choice, request.choice: resolved_mac_list}

    def _get_ether_resolved_mac(
        self, ip_objs, ether_gateway_map, ip_neighbors, choice
    ):
        arp_entries = {}
        for ip_obj in ip_objs:
            resolved_mac_list = ip_obj.ResolvedGatewayMac
            for index, gateway in enumerate(ip_obj.GatewayIp.Values):
                resolved_mac = resolved_mac_list[index]
                if re.search("unresolved", resolved_mac.lower()) is not None:
                    resolved_mac = None
                arp_entries[gateway] = resolved_mac

        ethernet_names = ip_neighbors.ethernet_names
        if ethernet_names is None:
            ethernet_names = ether_gateway_map.keys()
        resolved_mac_list = []
        for ethernet_name in ethernet_names:
            gateway_ips = ether_gateway_map[ethernet_name]
            for gateway_ip in gateway_ips:
                if gateway_ip not in arp_entries:
                    raise Exception(
                        "{} not found within current configured gateway ips".format(
                            gateway_ip
                        )
                    )
                if choice == "ipv4":
                    resolved_mac_list.append(
                        {
                            "ethernet_name": ethernet_name,
                            "ipv4_address": gateway_ip,
                            "link_layer_address": arp_entries[gateway_ip],
                        }
                    )
                elif choice == "ipv6":
                    resolved_mac_list.append(
                        {
                            "ethernet_name": ethernet_name,
                            "ipv6_address": gateway_ip,
                            "link_layer_address": arp_entries[gateway_ip],
                        }
                    )
        self.logger.debug(
            "These are resolved_mac_list: %s" % resolved_mac_list
        )
        return resolved_mac_list

    def _get_href(self, xpath):
        return xpath.replace("[", "/").replace("]", "")

    def select_properties(self, xpath, properties=[]):
        href = self._get_href(xpath)
        payload = {
            "selects": [
                {
                    "from": href,
                    "properties": properties,
                    "children": [],
                    "inlines": [
                        {
                            "child": "multivalue",
                            "properties": ["format", "pattern", "values"],
                        }
                    ],
                }
            ]
        }
        url = "%s/operations/select?xpath=true" % self.api._ixnetwork.href
        results = self.api._ixnetwork._connection._execute(url, payload)
        try:
            return results[0]
        except Exception:
            raise Exception("Problem to select %s" % href)

    def imports(self, imports):
        self.logger.debug("imports of portion of config")
        if len(imports) > 0:
            errata = self._resource_manager.ImportConfig(
                json.dumps(imports), False
            )
            for item in errata:
                self.api.warning(item)
            return len(errata) == 0
        return True

    def configure_value(self, source, attribute, value, enum_map=None):
        if value is None:
            return
        xpath = "/multivalue[@source = '{0} {1}']".format(source, attribute)
        if isinstance(value, list) and len(set(value)) == 1:
            value = value[0]
        if enum_map is not None:
            if isinstance(value, list):
                value = [enum_map[val] for val in value]
            else:
                value = enum_map[value]
        if isinstance(value, list):
            ixn_value = {
                "xpath": "{0}/valueList".format(xpath),
                "values": value,
            }
        else:
            ixn_value = {
                "xpath": "{0}/singleValue".format(xpath),
                "value": value,
            }
        return ixn_value



import time
import pytest


def wait_for_arp(snappi_api, max_attempts=10, poll_interval_sec=1):
    """
    Args:
    snappi_api: snappi api
    max_attempts: maximum attempts for timeout
    poll_interval_sec: interval poll second
    Return:
    returns number of attempts if arp is resolved within max attempts else fail
    """
    attempts = 0
    v4_gateway_macs_resolved = False
    v6_gateway_macs_resolved = False

    get_config = snappi_api.get_config()
    v4_addresses = []
    v6_addresses = []

    for device in get_config.devices:
        for ethernet in device.ethernets:
            for v4_address in ethernet.ipv4_addresses:
                v4_addresses.append(v4_address.address)
            for v6_address in ethernet.ipv6_addresses:
                v6_addresses.append(v6_address.address)

    while attempts < max_attempts:
        request = snappi_api.states_request()
        request.choice = request.IPV4_NEIGHBORS
        states = snappi_api.get_states(request)

        if len(v4_addresses) > 0:
            v4_link_layer_address = [
                state.link_layer_address
                for state in states.ipv4_neighbors
                if state.link_layer_address is not None
            ]
            if len(v4_addresses) == len(v4_link_layer_address):
                v4_gateway_macs_resolved = True
        else:
            v4_gateway_macs_resolved = True

        request = snappi_api.states_request()
        request.choice = request.IPV6_NEIGHBORS
        states = snappi_api.get_states(request)

        if len(v6_addresses) > 0:
            v6_link_layer_address = [
                state.link_layer_address
                for state in states.ipv6_neighbors
                if state.link_layer_address is not None
            ]
            if len(v6_addresses) == len(v6_link_layer_address):
                v6_gateway_macs_resolved = True
        else:
            v6_gateway_macs_resolved = True

        if v4_gateway_macs_resolved and v6_gateway_macs_resolved:
            break
        else:
            time.sleep(poll_interval_sec)
            attempts += 1

    print("Attempts: ", attempts)
    print("Maxmimum Attempts:", max_attempts)
    if attempts >= max_attempts:
        import pdb

        pdb.set_trace()
        raise Exception(
            "ARP is not resolved in {} seconds".format(
                max_attempts * poll_interval_sec
            )
        )

    return attempts


def static_lag(api, utils):
    """Demonstrates the following:
    1) Creating a lag comprised of multiple ports
    2) Creating emulated devices over the lag
    3) Creating traffic over the emulated devices that will transmit
    traffic to a single rx port.

        TX LAG              DUT             RX
        ------+         +---------+
        port 1|         |
        ..    | ------> |
        port n|         |
        ------+
    """
    config = api.config()
    p1, p2 = config.ports.port(
        name="txp1", location=utils.settings.ports[0]
    ).port(name="rxp2", location=utils.settings.ports[1])

    config.layer1.layer1(
        name="layer1",
        port_names=[p.name for p in config.ports],
        speed=utils.settings.speed,
        media=utils.settings.media,
    )

    lag1, lag2 = config.lags.lag(name="lag1").lag(name="lag2")
    lp1 = lag1.ports.port(port_name=p1.name)[-1]
    lp2 = lag2.ports.port(port_name=p2.name)[-1]
    lag1.protocol.static.lag_id = 1
    lag2.protocol.static.lag_id = 2

    lp1.ethernet.name, lp2.ethernet.name = "eth1", "eth2"

    lp1.ethernet.mac = "00:11:02:00:00:01"
    lp2.ethernet.mac = "00:22:02:00:00:01"

    lp1.ethernet.vlans.vlan(priority=1, name="vlan1", id=1)[-1]
    lp2.ethernet.vlans.vlan(priority=1, name="vlan2", id=1)[-1]

    packets = 2000
    f1_size = 74
    f2_size = 1500
    d1, d2 = config.devices.device(name="device1").device(name="device2")
    eth1, eth2 = d1.ethernets.add(), d2.ethernets.add()
    eth1.connection.port_name, eth2.connection.port_name = lag1.name, lag2.name
    eth1.name, eth2.name = "d_eth1", "d_eth2"
    eth1.mac, eth2.mac = "00:00:00:00:00:11", "00:00:00:00:00:22"
    ip1, ip2 = eth1.ipv4_addresses.add(), eth2.ipv4_addresses.add()
    ip1.name, ip2.name = "ip1", "ip2"
    ip1.address = "10.1.1.1"
    ip1.gateway = "10.1.1.2"
    ip2.address = "10.1.1.2"
    ip2.gateway = "10.1.1.1"
    f1, f2 = config.flows.flow(name="f1").flow(name="f2")
    f1.tx_rx.port.tx_name = p1.name
    f1.tx_rx.port.rx_name = p2.name
    f2.tx_rx.port.rx_name = p1.name
    f2.tx_rx.port.tx_name = p2.name
    config.options.port_options.location_preemption = True
    f1.duration.fixed_packets.packets = packets
    f2.duration.fixed_packets.packets = packets
    f1.size.fixed = f1_size
    f2.size.fixed = f2_size
    f1.rate.percentage = 10
    f2.rate.percentage = 10

    f1.metrics.enable = True
    f1.metrics.loss = True

    f2.metrics.enable = True
    f2.metrics.loss = True

    api.set_config(config)

    wait_for_arp(api, max_attempts=10, poll_interval_sec=2)

    print("Starting transmit on all flows ...")
    cs = api.control_state()
    cs.traffic.flow_transmit.state = cs.traffic.flow_transmit.START
    api.set_control_state(cs)

    utils.wait_for(lambda: utils.is_traffic_stopped(api), "traffic to stop")

    utils.wait_for(
        lambda: utils.is_stats_accumulated(api, packets * 2),
        "stats to be accumulated",
    )

    utils.wait_for(
        lambda: results_ok(api, utils, f1_size, f2_size, packets),
        "stats to be as expected",
        timeout_seconds=30,
    )


def results_ok(api, utils, size1, size2, packets):
    """
    Returns true if stats are as expected, false otherwise.
    """
    port_results, flow_results = utils.get_all_stats(api)
    frames_ok = utils.total_frames_ok(port_results, flow_results, packets * 2)
    bytes_ok = utils.total_bytes_ok(
        port_results, flow_results, packets * size1 + packets * size2
    )
    return frames_ok and bytes_ok

@pytest.mark.skip(reason="skip to CICD faster")
def test_static_lag(api, utils):
    for i in range(0, 4):
        static_lag(api, utils)
    # test1(api, utils)
    # test2(api, utils)



import pytest


@pytest.mark.skip(reason="Not yet implemented")
def test_get_ingress_results(b2b_raw_config, options, tx_port, rx_port, api):
    """UDP Flow test traffic configuration"""
    import snappi

    b2b_raw_config = snappi.Api().config()
    b2b_raw_config.flows.clear()
    f = b2b_raw_config.flows.flow()[-1]
    f.name = "UDP Flow"
    f.packet.ethernet().vlan().ipv4().udp()
    f.tx_rx.port.tx_name = tx_port.name
    f.tx_rx.port.rx_name = rx_port.name
    f.size.fixed = 128
    f.rate.pps = 1000
    f.duration.fixed_packets.packets = 10000
    udp = f.packet[-1]
    udp.src_port.increment.start = 12001
    udp.dst_port.value = 20
    api.set_config(b2b_raw_config)
    # udp.ingress_result_name = 'UDP SRC PORT'
    transmit_state = api.transmit_state()
    transmit_state.state = "start"
    api.set_transmit_state(transmit_state)

    request = api.metrics_request()
    request.choice = request.FLOW
    request.flow.ingress_result_names = ["UDP SRC PORT"]
    while True:
        df = api.get_metrics(request).flow_metrics
        if df.frames_tx.sum() >= 10000 and df.frames_tx_rate.sum() == 0:
            break


if __name__ == "__main__":
    pytest.main(["-s", __file__])



import pytest

@pytest.mark.skip(reason="CI failure")
# @pytest.fixture
def port_configs(api, utils):
    """This fixture demonstrates setting up configurations that consist
    only of port, layer1 and device settings.
    """
    configs = []
    for c in range(2):
        config = api.config()
        ports = config.ports.port(
            name="Port 1", location=utils.settings.ports[0]
        ).port(name="Port 2", location=utils.settings.ports[1])
        config.options.port_options.location_preemption = True
        config.layer1.layer1().layer1()
        devices = config.devices.device().device()
        for i, l1 in enumerate(config.layer1):
            l1.name = "L1 Settings %s" % (i)
            l1.port_names = [ports[i].name]
            l1.speed = utils.settings.speed
            l1.flow_control.ieee_802_1qbb.pfc_delay = 1
            l1.flow_control.ieee_802_1qbb.pfc_class_0 = 0
            l1.flow_control.ieee_802_1qbb.pfc_class_1 = 1
            l1.flow_control.ieee_802_1qbb.pfc_class_2 = 2
            l1.flow_control.ieee_802_1qbb.pfc_class_3 = 3
            l1.flow_control.ieee_802_1qbb.pfc_class_4 = 4
            l1.flow_control.ieee_802_1qbb.pfc_class_5 = 5
            l1.flow_control.ieee_802_1qbb.pfc_class_6 = 6
            l1.flow_control.ieee_802_1qbb.pfc_class_7 = 7
            devices[i].name = "Device %s" % (i)
            eth = devices[i].ethernets.add()
            eth.connection.port_name = ports[i].name
            eth.name = "Ethernet %s" % (i)
            eth.mac = "00:00:00:00:00:{:02x}".format(i)
            ip = eth.ipv4_addresses.add()
            ip.name = "Ipv4 %s" % (i)
            ip.gateway = "1.1.1.2"
            ip.address = "1.1.1.1"
        configs.append(config)
    return configs

@pytest.mark.skip(reason="CI failure")
# @pytest.fixture
def flow_configs(port_configs):
    """This fixture demonstrates adding flows to port configurations."""

    for config in port_configs:
        f = config.flows.flow()[-1]
        f.tx_rx.device.tx_names = [config.devices[0].name]
        f.tx_rx.device.rx_names = [config.devices[1].name]
        f.name = "%s --> %s" % (config.ports[0].name, config.ports[1].name)
        f.size.fixed = 128
        f.duration.fixed_packets.packets = 10000000
    return port_configs

@pytest.mark.skip(reason="CI failure")
def test_fixtures(flow_configs, api):
    """Iterate through the flow configs using each config to run a test."""
    for config in flow_configs:
        api.set_config(config)



import pytest
import logging
import snappi
import utils as utl


def test_mac_addrs(b2b_raw_config_vports):
    """create snappi api with enable loglevel to logging.DEBUG
    Validate loglevel is satting with proper value"""
    api = snappi.api(
        location=utl.settings.location,
        ext=utl.settings.ext,
        loglevel=logging.DEBUG,
    )
    utl.configure_credentials(api, utl.settings.username, utl.settings.psd)

    api.set_config(b2b_raw_config_vports)
    assert api.log_level == logging.DEBUG


if __name__ == "__main__":
    pytest.main(["-s", __file__])



import pytest


@pytest.mark.skip(
    reason="""
    Skipping as this test as starting the protcols,
    might cause an issue in CI/CD
    """
)
def test_bgp_speaker(api, b2b_raw_config):
    """This is the configuration steps in https://github.com/Azure/sonic-mgmt/blob/master/tests/bgp/test_bgp_speaker.py

    logging.info("Start exabgp on ptf")
    for i in range(0, 3):
        local_ip = str(speaker_ips[i].ip)
        ptfhost.exabgp(name="bgps%d" % i,
                       state="started",
                       local_ip=local_ip,
                       router_id=local_ip,
                       peer_ip=lo_addr,
                       local_asn=bgp_speaker_asn,
                       peer_asn=mg_facts['minigraph_bgp_asn'],
                       port=str(port_num[i]))
    """
    local_ip = ["10.10.0.1", "20.20.0.1", "30.30.0.1"]
    local_asn = "65000"
    peer_asn = "65000"

    dev = b2b_raw_config.devices.device()[-1]
    dev.name = "BGP device"
    dev.container_name = b2b_raw_config.ports[0].name
    eth = dev.ethernet
    ipv4 = eth.ipv4
    ipv4.name = "ipv4"
    ipv4.address.values = local_ip
    ipv4.gateway.value = peer_ip
    bgpv4 = ipv4.bgpv4
    bgpv4.name = "bgpv4"
    bgpv4.router_id.values, bgpv4.dut_ipv4_address.value = local_ip, peer_ip
    bgpv4.as_number.value, bgpv4.dut_as_number.value = local_asn, peer_asn
    b2b_raw_config.flows.clear()
    api.set_config(b2b_raw_config)
    transmit_state = api.transmit_state()
    transmit_state.state = "start"
    api.set_transmit_state(transmit_state)
    import time

    time.sleep(5)
    transmit_state.state = "stop"
    api.set_transmit_state(transmit_state)



from snappi_ixnetwork.device.base import Base
from snappi_ixnetwork.logger import get_ixnet_logger


class Isis(Base):
    _NETWORK_TYPE = {
        "network_type": {
            "ixn_attr": "networkType",
            "default_value": "broadcast",
            "enum_map": {
                "broadcast": "broadcast",
                "point_to_point": "pointpoint"
            },
        },
    }

    _LEVEL_TYPE = {
        "level_type": {
            "ixn_attr": "levelType",
            "default_value": "level_2",
            "enum_map": {
                "level_1": "level1",
                "level_2": "level2",
                "level_1_2": "l1l2"
            },
        },
    }

    _AUTH_TYPE = {
        "auth_type": {
            "ixn_attr": "authType",
            "enum_map": {"md5": "md5", "password": "password"},
        },
        "mde5": "md5",
        "password": "password",
    }

    _ORIGIN_TYPE = {
        "origin_type": {
            "ixn_attr": "routeOrigin",
            "default_value": "internal",
            "enum_map": {"internal": "internal", "external": "external"},
        }
    }

    _REDISTRIBUTION_TYPE = {
        "redistribution_type": {
            "ixn_attr": "redistribution",
            "default_value": "up",
            "enum_map": {"up": "up", "down": "down"},
        }
    }

    _BASIC = {
        "ipv4_te_router_id": "tERouterId",
        "hostname": "hostName",
        "enable_wide_metric": "enableWideMetric",
        "learned_lsp_filter": "discardLSPs",
    }

    _ADVANCED = {
        "enable_hello_padding": "enableHelloPadding",
        "max_area_addresses": "maxAreaAddresses",
        "area_addresses": "areaAddresses",
        "lsp_refresh_rate": "lSPRefreshRate",
        "lsp_lifetime": "lSPLifetime",
        "psnp_interval": "pSNPInterval",
        "csnp_interval": "cSNPInterval",
        "max_lsp_size": "maxLSPSize",
        "lsp_mgroup_min_trans_interval": "lSPorMGroupPDUMinTransmissionInterval",   # noqa
        "enable_attached_bit": "attached",
    }

    _L1_SETTINGS = {
        "priority": "level1Priority",
        "hello_interval": "level1HelloInterval",
        "dead_interval": "level1DeadInterval",
    }

    _L2_SETTINGS = {
        "priority": "level2Priority",
        "hello_interval": "level2HelloInterval",
        "dead_interval": "level2DeadInterval",
    }

    _MULTI_TOPOLOGY_IDS = {
        "mt_id": "mtId",
        "link_metric": "linkMetric",
    }

    _TRAFFIC_ENGINEERING = {
        "administrative_group": "administratorGroup",
        "metric_level": "metricLevel",
        "max_bandwith": "maxBandwidth",
        "max_reservable_bandwidth": "maxReservableBandwidth",
    }

    _PRIORITY_BANDWIDTHS = {
        "pb0": "bandwidthPriority0",
        "pb1": "bandwidthPriority1",
        "pb2": "bandwidthPriority2",
        "pb3": "bandwidthPriority3",
        "pb4": "bandwidthPriority4",
        "pb5": "bandwidthPriority5",
        "pb6": "bandwidthPriority6",
        "pb7": "bandwidthPriority7",
    }

    _ADVANCED_INTERFACE = {
        "auto_adjust_mtu": "autoAdjustMTU",
        "auto_adjust_area": "autoAdjustArea",
        "auto_adjust_supported_protocols": "autoAdjustSupportedProtocols",
        "enable_3way_handshake": "enable3WayHandshake",
    }

    _LINK_PROTECTION = {
        "extra_traffic": "extraTraffic",
        "unprotected": "unprotected",
        "shared": "shared",
        "dedicated_1_to_1": "dedicatedOneToOne",
        "dedicated_1_plus_1": "dedicatedOnePlusOne",
        "enhanced": "enhanced",
        "reserved_40": "reserved0x40",
        "reserved_80": "reserved0x80",
    }
    _ROUTER_AUTH = {
        "ignore_receive_md5": "ignoreReceiveMD5",
        "area_auth": {
            "ixn_attr": "areaAuthenticationType",
            "enum_map": {"md5": "areaTransmitPasswordOrMD5Key", "password": "areaTransmitPasswordOrMD5Key"},    # noqa
        },
        "domain_auth": {
            "ixn_attr": "domainAuthenticationType",
            "enum_map": {"md5": "domainTransmitPasswordOrMD5Key", "password": "domainTransmitPasswordOrMD5Key"},    # noqa
        },
    }

    _IP_POOL = {
        "address": "networkAddress",
        "prefix": "prefixLength",
        "count": "numberOfAddressesAsy",
        "step": "prefixAddrStep",
    }
    
    def __init__(self, ngpf):
        super(Isis, self).__init__()
        self._ngpf = ngpf
        self.logger = get_ixnet_logger(__name__)
        self._system_id = None
        self._isis_router_name = None
        self._isis_interface_name = None

    def config(self, device):
        self.logger.debug("Configuring ISIS")
        isis = device.get("isis")
        if isis is None:
            return
        self._system_id = isis.get("system_id")
        self._isis_router_name = isis.get("name")
        self._add_isis_router(isis)

    def _add_isis_router(self, isis):
        self.logger.debug("Configuring Isis Router")
        interfaces = isis.get("interfaces")
        if interfaces is None:
            return
        # IxNetwork supports single ISIS interface per router
        if len(interfaces) > 1:
            return
        for interface in interfaces:
            ethernet_name = interface.get("eth_name")
            self._ngpf.working_dg = self._ngpf.api.ixn_objects.get_working_dg(
                ethernet_name
            )
            if not self._is_valid(ethernet_name):
                continue
            ixn_eth = self._ngpf.api.ixn_objects.get_object(ethernet_name)
            ixn_isis = self.create_node_elemet(
                ixn_eth, "isisL3", interface.get("name")
            )
            self._ngpf.set_device_info(interface, ixn_isis)
            self._config_isis_interface(interface, ixn_isis)
            ixn_isis_router = self.create_node_elemet(
                self._ngpf.working_dg, "isisL3Router", isis.get("name")
            )
            ixn_bridged_data = self.create_node_elemet(
                self._ngpf.working_dg, "bridgeData" 
            )
            self._config_system_id(isis, ixn_bridged_data)
            self._config_isis_router(isis, ixn_isis_router)
            self._add_isis_route_range(isis, ixn_isis_router, ixn_isis)
            
    def _is_valid(self, ethernet_name):
        is_valid = True
        if is_valid:
            self.logger.debug("Isis validation success")
        else:
            self.logger.debug("Isis validation failure")
        return is_valid

    def _config_system_id(self, isis, ixn_bridged_data):
        system_id = isis.get("system_id")
        ixn_bridged_data["systemId"] = self.multivalue(system_id)

    def _config_isis_interface(self, interface, ixn_isis):
        self.logger.debug("Configuring Isis interfaces")        
        # Metric
        metric = interface.get("metric")
        ixn_isis["interfaceMetric"] = self.multivalue(metric)
        # Network Type
        network_type = interface.get("network_type")
        mapped_type = Isis._NETWORK_TYPE["network_type"]["enum_map"][network_type]   # noqa
        ixn_isis["networkType"] = self.multivalue(mapped_type)
        # Level Type
        level_type = interface.get("level_type")
        mapped_level = Isis._LEVEL_TYPE["level_type"]["enum_map"][level_type]   # noqa
        ixn_isis["levelType"] = self.multivalue(mapped_level)
        # L1 Settings
        l1_settings = interface.get("l1_settings")
        if l1_settings is not None:
            self.logger.debug("priority %s hello_interval %s dead_interval %s " % (l1_settings.priority, l1_settings.hello_interval, l1_settings.dead_interval)) # noqa
            self.configure_multivalues(l1_settings, ixn_isis, Isis._L1_SETTINGS)  # noqa
        # L2 Settings
        l2_settings = interface.get("l2_settings")
        if l2_settings is not None:
            self.logger.debug("priority %s hello_interval %s dead_interval %s " % (l2_settings.priority, l2_settings.hello_interval, l2_settings.dead_interval)) # noqa
            self.configure_multivalues(l2_settings, ixn_isis, Isis._L2_SETTINGS)  # noqa
        # Multiple Topology IDs
        self._configure_multi_topo_id(interface, ixn_isis)
        # Traffic Engineering
        self._configure_traffic_engineering(interface, ixn_isis)
        # Authentication
        auth = interface.get("authentication")
        if auth is not None:
            self.logger.debug("authentication %s " % (auth.auth_type))
            self.configure_multivalues(auth, ixn_isis, Isis._AUTH_TYPE)
        # Advanced
        advanced = interface.get("advanced")
        if advanced is not None:
            self.logger.debug("auto_adjust_mtu %s auto_adjust_area %s auto_adjust_supported_protocols %s enable_3way_handshake %s p2p_hellos_to_unicast_mac %s " % (advanced.auto_adjust_mtu, advanced.auto_adjust_area, advanced.auto_adjust_supported_protocols, advanced.enable_3way_handshake, advanced.p2p_hellos_to_unicast_mac)) # noqa
            self.configure_multivalues(advanced, ixn_isis, Isis._ADVANCED_INTERFACE) # noqa
        # Link Protection
        link_protection = interface.get("link_protection")
        if link_protection is not None:
            self.logger.debug("Configuring link protection")
            self.configure_multivalues(link_protection, ixn_isis, Isis._LINK_PROTECTION) # noqa
        # srlg values
        srlg_vals = interface.get("srlg_values")
        if srlg_vals is not None:
            srlg_count = len(srlg_vals)
            if srlg_count > 0:
                self.logger.debug("srlg values")
                ixn_isis["enableSRLG"] = True
                ixn_isis["srlgCount"] = srlg_count
                for index, value in enumerate(srlg_vals):
                    ixn_isis["srlgValueList"][index] = self.multivalue(value)
        #TBD 
        # Adjacency Sids
        self._configure_adjacency_sids(interface, ixn_isis)     

    # TBD 
    def _configure_multi_topo_id(self, interface, ixn_isis):
        "Configuring multiple topology IDs"
    
    # TBD 
    def _configure_traffic_engineering(self, interface, ixn_isis):
        "Configuring Traffic Engineering"

    # TBD 
    def _configure_adjacency_sids(self, interface, ixn_isis):
        "Configuring Adjacency sids"  

    def _config_isis_router(self, otg_isis_router, ixn_isis_router):
        "Configuring Isis router"
        isis_router_basic = otg_isis_router.get("basic")
        if isis_router_basic is not None:
            self._configure_isis_router_basic(isis_router_basic, ixn_isis_router) # noqa
        isis_router_advanced = otg_isis_router.get("advanced")
        if isis_router_advanced is not None:
            self._configure_isis_router_advanced(isis_router_advanced, ixn_isis_router) # noqa
        isis_router_auth = otg_isis_router.get("router_auth")
        if isis_router_auth is not None:
            self._configure_isis_router_auth(isis_router_auth, ixn_isis_router) # noqa
        
    def _configure_isis_router_basic(self, otg_router_basic, ixn_isis_router):
        "Configuring ISIS router basic"
        ipv4_te_router_id = otg_router_basic.get("ipv4_te_router_id")
        if ipv4_te_router_id is not None:
            ixn_isis_router["enableTE"] = self.multivalue(True)
            ixn_isis_router["tERouterId"] = self.multivalue(ipv4_te_router_id)
        else:
            ixn_isis_router["enableTE"] = self.multivalue(False)
        hostname = otg_router_basic.get("hostname")
        if hostname is not None:
            ixn_isis_router["enableHostName"] = self.multivalue(True)
            ixn_isis_router["hostName"] = self.multivalue(hostname)
        else:
            ixn_isis_router["enableHostName"] = self.multivalue(False) 
        ixn_isis_router["discardLSPs"] = self.multivalue(otg_router_basic.get("learned_lsp_filter")) # noqa
        ixn_isis_router["enableWideMetric"] = self.multivalue(otg_router_basic.get("enable_wide_metric")) # noqa

    def _configure_isis_router_advanced(self, otg_router_advanced, ixn_isis_router): # noqa
        "Configuring ISIS router advanced"
        ixn_isis_router["enableHelloPadding"] = self.multivalue(otg_router_advanced.get("enable_hello_padding")) # noqa
        ixn_isis_router["maxAreaAddresses"] = self.multivalue(otg_router_advanced.get("max_area_addresses")) # noqa
        area_address = "490001"
        if len(otg_router_advanced.get("area_addresses")) > 0:
            area_address = "".join(otg_router_advanced.get("area_addresses"))
        ixn_isis_router["areaAddresses"] = self.multivalue(area_address) # noqa
        ixn_isis_router["lSPRefreshRate"] = self.multivalue(otg_router_advanced.get("lsp_refresh_rate")) # noqa
        ixn_isis_router["lSPLifetime"] = self.multivalue(otg_router_advanced.get("lsp_lifetime")) # noqa
        ixn_isis_router["pSNPInterval"] = self.multivalue(otg_router_advanced.get("psnp_interval")) # noqa
        ixn_isis_router["cSNPInterval"] = self.multivalue(otg_router_advanced.get("csnp_interval")) # noqa
        ixn_isis_router["maxLSPSize"] = self.multivalue(otg_router_advanced.get("max_lsp_size")) # noqa
        ixn_isis_router["lSPorMGroupPDUMinTransmissionInterval"] = self.multivalue(otg_router_advanced.get("lsp_mgroup_min_trans_interval")) # noqa
        ixn_isis_router["attached"] = self.multivalue(otg_router_advanced.get("enable_attached_bit")) # noqa
        
    def _configure_isis_router_auth(self, otg_router_auth, ixn_isis_router): # noqa
        "Configuring ISIS router authentication"
        self.configure_multivalues(otg_router_auth, ixn_isis_router, Isis._ROUTER_AUTH) # noqa

    def _add_isis_route_range(self, otg_isis_router, ixn_isis_router, ixn_isis):
        "Configuring ISIS route range"
        v4_routes = otg_isis_router.get("v4_routes")
        if v4_routes is not None:
            self._configure_isisv4_route(v4_routes, ixn_isis)
        v6_routes = otg_isis_router.get("v6_routes")
        if v6_routes is not None:
            self._configure_isisv6_route(v6_routes, ixn_isis)
        self._ngpf.compactor.compact(self._ngpf.working_dg.get("networkGroup"))
        
    def _configure_isisv4_route(self, v4_routes, ixn_isis):
        "Configuring ISIS v4 routes"
        if v4_routes is None:
            return
        self.logger.debug("Configuring ISISv4 Route")
        for route in v4_routes:
            addresses = route.get("addresses")
            for address in addresses:
                ixn_ng = self.create_node_elemet(
                    self._ngpf.working_dg, "networkGroup", route.get("name")
                )
                ixn_ng["multiplier"] = 1
                ixn_ip_pool = self.create_node_elemet(
                    ixn_ng, "ipv4PrefixPools", route.get("name")
                )
                ixn_connector = self.create_property(ixn_ip_pool, "connector")
                ixn_connector["connectedTo"] = self.post_calculated(
                    "connectedTo", ref_ixnobj=ixn_isis
                )
                self.configure_multivalues(address, ixn_ip_pool, Isis._IP_POOL)
                ixn_route = self.create_node_elemet(
                    ixn_ip_pool, "isisL3RouteProperty", route.get("name")
                )
                self._ngpf.set_device_info(route, ixn_ip_pool)
                self._configure_route(route, ixn_route)

    def _configure_isisv6_route(self, v6_routes, ixn_isis):
        "Configuring ISIS v6 routes"
        if v6_routes is None:
            return
        self.logger.debug("Configuring ISISv6 Route")
        for route in v6_routes:
            addresses = route.get("addresses")
            for address in addresses:
                ixn_ng = self.create_node_elemet(
                    self._ngpf.working_dg, "networkGroup", route.get("name")
                )
                ixn_ng["multiplier"] = 1
                ixn_ip_pool = self.create_node_elemet(
                    ixn_ng, "ipv6PrefixPools", route.get("name")
                )
                ixn_connector = self.create_property(ixn_ip_pool, "connector")
                ixn_connector["connectedTo"] = self.post_calculated(
                    "connectedTo", ref_ixnobj=ixn_isis
                )
                self.configure_multivalues(address, ixn_ip_pool, Isis._IP_POOL)
                ixn_route = self.create_node_elemet(
                    ixn_ip_pool, "isisL3RouteProperty", route.get("name")
                )
                self._ngpf.set_device_info(route, ixn_ip_pool)
                self._configure_route(route, ixn_route)
        
    def _configure_route(self, otg_route, ixn_route):
        "Configuring ISIS v4 routes"
        self._ngpf.set_ixn_routes(otg_route, ixn_route)
        # Link metric
        metric = otg_route.get("link_metric")
        ixn_route["metric"] = self.multivalue(metric)
        # Origin Type
        origin_type = otg_route.get("origin_type")
        mapped_type = Isis._ORIGIN_TYPE["origin_type"]["enum_map"][origin_type]   # noqa
        ixn_route["routeOrigin"] = self.multivalue(mapped_type)
        # Redistribution Type
        redistribution_type = otg_route.get("redistribution_type")
        mapped_type = Isis._REDISTRIBUTION_TYPE["redistribution_type"]["enum_map"][redistribution_type]   # noqa
        ixn_route["redistribution"] = self.multivalue(mapped_type)


class Validation(object):
    """Validate the configuration

    Ensures entire configuration has unique names

    Args
    ----
    - ixnetworkapi (Api): instance of the ixnetworkapi class
    """

    def __init__(self, ixnetworkapi):
        self._api = ixnetworkapi

    def validate_config(self):
        self._unique_name_errors = []
        self.__check_config_objects(self._api.snappi_config, item_ids=[])
        if len(self._unique_name_errors) > 0:
            raise NameError(", ".join(self._unique_name_errors))

    def __check_config_objects(self, config_item, item_ids):
        if config_item is None:
            return
        config_item.validate()

        if (
            hasattr(config_item, "choice") is True
            and getattr(config_item, "choice") is None
        ):
            return

        for attr_name in config_item._properties:
            if attr_name.startswith("_") or attr_name == "parent":
                continue
            attr_value = config_item.get(attr_name)
            if callable(attr_value) is True:
                continue

            if attr_name == "name":
                if attr_value in self._api._config_objects:
                    self._unique_name_errors.append(
                        '%s.name: "%s" is not unique'
                        % (config_item.__class__.__name__, attr_value)
                    )
                # todo: we will enable after finalyze model
                if attr_value is None:
                    continue
                    # self._unique_name_errors.append('%s.name: "None" is not allowed' % (config_item.__class__.__name__))
                else:
                    self._api._config_objects[attr_value] = config_item
            elif "__module__" in dir(attr_value):
                if attr_value.__module__.startswith("snappi"):
                    if "__next__" in dir(attr_value):
                        for item in attr_value:
                            item_id = id(item)
                            if item_id in item_ids:
                                continue
                            if getattr(item, "parent", None) is not None:
                                item_ids.append(item_id)
                                self.__check_config_objects(
                                    item.parent, item_ids
                                )
                            else:
                                self.__check_config_objects(item, item_ids)
                    else:
                        self.__check_config_objects(attr_value, item_ids)



from ixnetwork_restpy import errors as err
import sys
import traceback


class SnappiIxnException(Exception):
    def __init__(self, *args):
        super(SnappiIxnException, self).__init__(*args)
        self._args = args
        self._message = None
        self._status_code = None
        self.process_exception()
        self._add_traceback()

    @property
    def args(self):
        return (
            self._status_code,
            (
                [self._message]
                if not isinstance(self._message, list)
                else self._message
            ),
        )

    @property
    def message(self):
        return (
            [self._message]
            if not isinstance(self._message, list)
            else self._message
        )

    @property
    def status_code(self):
        return self._status_code

    def process_exception(self):
        if isinstance(self._args, tuple) and len(self._args) == 1:
            if isinstance(self._args[0], (str, list)):
                self._status_code = (
                    500 if self._status_code is None else self._status_code
                )
                self._message = self._args[0]
                return
            if isinstance(self._args[0], err.IxNetworkError):
                self._status_code = self._args[0].status_code
                self._status_code = (
                    500 if self._status_code is None else self._status_code
                )
                self._message = self._args[0].message
                return
            if isinstance(self._args[0], (NameError, TypeError, ValueError)):
                self._status_code = 400
                self._args = self._args[0].args
                return self.process_exception()
            if isinstance(
                self._args[0], (ImportError, Exception, RuntimeError)
            ):
                self._args = self._args[0].args
                return self.process_exception()
        elif isinstance(self._args, tuple) and len(self._args) > 1:
            self._status_code = self._args[0]
            self._message = self._args[1]
            return
        else:
            self._message = self._args
        return

    def _add_traceback(self):
        tb = sys.exc_info()
        if len(tb) >= 2:
            tb = traceback.format_tb(tb[2])
            if isinstance(tb, str):
                tb = [tb]
            if isinstance(self._message, list):
                tb.extend(self._message)
                self._message = tb
            if isinstance(self._message, str):
                self._message = "{} {}".format("".join(tb), self._message)

    def __str__(self):
        if isinstance(self._message, list):
            return "".join(self._message)
        return self._message

    def __repr__(self):
        if isinstance(self._message, list):
            return "".join(self._message)
        return self._message



def test_phb_ecn(api, tx_port, rx_port, b2b_raw_config):
    """
    This will test that phb and ecn are set on an ipv4 header
    """
    b2b_raw_config.flows.clear()
    f = b2b_raw_config.flows.flow()[-1]
    f.name = "Ipv4 with Phb and Ecn"
    f.tx_rx.port.tx_name = tx_port.name
    f.tx_rx.port.rx_name = rx_port.name
    f.packet.ethernet().ipv4()
    ip = f.packet[-1]
    ip.priority.choice = ip.priority.DSCP
    ip.priority.dscp.phb.values = [
        ip.priority.dscp.phb.CS2,
        ip.priority.dscp.phb.CS1,
        ip.priority.dscp.phb.CS5,
    ]
    ip.priority.dscp.ecn.value = ip.priority.dscp.ecn.CAPABLE_TRANSPORT_1
    api.set_config(b2b_raw_config)



from snappi_ixnetwork.device.base import Base
from snappi_ixnetwork.logger import get_ixnet_logger


class LoopbackInt(Base):
    def __init__(self, ngpf):
        super(LoopbackInt, self).__init__()
        self._ngpf = ngpf
        self._ixn_parent_dgs = []
        self.logger = get_ixnet_logger(__name__)

    def config(self):
        self._ixn_parent_dgs = []
        for device in self._ngpf.api.snappi_config.devices:
            vxlan_source_int_list = self._get_vxlan_source_ints(device)
            ipv4_loopbacks = device.get("ipv4_loopbacks")
            if ipv4_loopbacks is not None:
                self._config_ipv4_loopbacks(
                    ipv4_loopbacks, device, vxlan_source_int_list
                )
            ipv6_loopbacks = device.get("ipv6_loopbacks")
            if ipv6_loopbacks is not None:
                self._config_ipv6_loopbacks(
                    ipv6_loopbacks, device, vxlan_source_int_list
                )
        return self._ixn_parent_dgs

    def _get_vxlan_source_ints(self, device):
        vxlan_source_int_list = []
        vxlan = device.get("vxlan")
        if vxlan is not None:
            v4_tunnels = vxlan.get("v4_tunnels")
            if v4_tunnels is not None and len(v4_tunnels) > 0:
                for v4_tunnel in v4_tunnels:
                    vxlan_source_int_list.append(
                        v4_tunnel.get("source_interface")
                    )

            v6_tunnels = vxlan.get("v6_tunnels")
            if v6_tunnels is not None and len(v6_tunnels) > 0:
                for v6_tunnel in v6_tunnels:
                    vxlan_source_int_list.append(
                        v6_tunnel.get("source_interface")
                    )
        return vxlan_source_int_list

    def _create_dg(self, loop_back, device):
        self.logger.debug("Configuring DG for loopback interface")
        eth_name = loop_back.get("eth_name")
        if eth_name not in self._ngpf.api.ixn_objects.names:
            raise Exception(
                "Ethernet %s not present within configuration" % eth_name
            )
        ixn_parent_dg = self._ngpf.api.ixn_objects.get_working_dg(eth_name)
        self._ixn_parent_dgs.append(ixn_parent_dg)
        ixn_dg = self.create_node_elemet(
            ixn_parent_dg,
            "deviceGroup",
            "loopback_{}".format(device.get("name")),
        )
        ixn_dg["multiplier"] = 1
        self._ngpf.working_dg = ixn_dg
        self._ngpf.set_device_info(device, ixn_dg)
        return ixn_dg

    def _config_ipv4_loopbacks(
        self, ipv4_loopbacks, device, vxlan_source_int_list
    ):
        self.logger.debug("Configuring IPv4 loopback interface")
        for ipv4_loopback in ipv4_loopbacks:
            ixn_dg = self._create_dg(ipv4_loopback, device)
            name = ipv4_loopback.get("name")
            if name in vxlan_source_int_list:
                ixn_eth = self.create_node_elemet(
                    ixn_dg, "ethernet", "eth {}".format(name)
                )
                ixn_v4 = self.create_node_elemet(ixn_eth, "ipv4", name)
                self._ngpf.set_device_info(ipv4_loopback, ixn_v4)
                ixn_v4["address"] = self.as_multivalue(
                    ipv4_loopback, "address"
                )
            else:
                ixn_v4lb = self.create_node_elemet(
                    ixn_dg, "ipv4Loopback", name
                )
                self._ngpf.set_device_info(ipv4_loopback, ixn_v4lb)
                ixn_v4lb["address"] = self.as_multivalue(
                    ipv4_loopback, "address"
                )

    def _config_ipv6_loopbacks(
        self, ipv6_loopbacks, device, vxlan_source_int_list
    ):
        self.logger.debug("Configuring IPv6 loopback interface")
        for ipv6_loopback in ipv6_loopbacks:
            ixn_dg = self._create_dg(ipv6_loopback, device)
            name = ipv6_loopback.get("name")
            if name in vxlan_source_int_list:
                ixn_eth = self.create_node_elemet(
                    ixn_dg, "ethernet", "eth {}".format(name)
                )
                ixn_v6 = self.create_node_elemet(ixn_eth, "ipv6", name)
                self._ngpf.set_device_info(ipv6_loopback, ixn_v6)
                ixn_v6["address"] = self.as_multivalue(
                    ipv6_loopback, "address"
                )
            else:
                ixn_v4lb = self.create_node_elemet(
                    ixn_dg, "ipv6Loopback", ipv6_loopback.get("name")
                )
                self._ngpf.set_device_info(ipv6_loopback, ixn_v4lb)
                ixn_v4lb["address"] = self.as_multivalue(
                    ipv6_loopback, "address"
                )
